#include <stdarg.h>
#include <stdio.h>
#include "CIRC.h"
#include "Brains.h"
#include "MsgQueue.h"

CLinkedList<Brain *> Brains;

extern char VerboseNicks[256];
extern Brain    Default;
extern DICTIONARY *Words;
extern void GetLine(char *line, int len);
extern void DoCommaSep(char *str, void (* Funct)(char *));
extern int ThinkTime;
extern float SurpriseAvg;

void ShowBrain(Brain *brain)
{
   int i, len;

   putchar('\n');
   printf(" %s\n", brain->FileName);
   putchar(' ');
   len = strlen(brain->FileName);
   for(i=0;i<len;i++) putchar('-');
   putchar('\n');

   printf(" IsMute: %d  VerbLog: %d  Log: %d  CanLearn: %d  WriteTrn: %d\n"
          " ThinkTime: %d  SurpriseAvg: %f\n",
          brain->IsMute, brain->VerbLog, brain->Log, brain->CanLearn, brain->WriteTrn,
          brain->ThinkTime, brain->SurpriseAvg);
   len = brain->Chans.NumNodes();
   
   puts("\n Attached:\n ---------");
   for(i=0;i<len;i++) printf(" %s\n", brain->Chans.GetDatabyIndex(i));

   puts("\n Ignored:\n --------");
   len = brain->Ignores.NumNodes();
   for(i=0;i<len;i++) printf(" %s\n", brain->Ignores.GetDatabyIndex(i));
   putchar('\n');
}


void Attach(Brain *brain, char *chan)
{
   char *seps  = " ,";
   char *token = strtok(chan, seps);

   while(token)
   {
      AttachBrain(brain, token);
      token = strtok(NULL, seps);
   }
}


void Detach(char *chan)
{
   DoCommaSep(chan, DetachBrain);
}


Brain * FindBrain(char *chan)
{
   int i,j,len,nodes;
   Brain *brain;
   
   len = Brains.NumNodes();
   for(i=0;i<len;i++)
   {
      brain = Brains.GetDatabyIndex(i);
      nodes = brain->Chans.NumNodes();
      for(j=0;j<nodes;j++) if (!strcmp(chan, brain->Chans.GetDatabyIndex(j))) return brain;
   }

   brain = &Default;
   nodes = brain->Chans.NumNodes();
   for(j=0;j<nodes;j++) if (!strcmp(chan, brain->Chans.GetDatabyIndex(j))) return brain;
   
   return NULL;
}


void AttachBrain(Brain *brain, char *chan)
{
   int len;
   char *temp;

   len = strlen(chan);
   strlwr(chan);

   while(FindBrain(chan)) DetachBrain(chan);

   temp = new char[len+1];
   strcpy(temp, chan);
   brain->Chans.Add(temp);
}


void DetachBrain(char *chan)
{
   int i=0,len;
   Brain *brain;
   
   if (chan)
   {
      strlwr(chan);
      brain = FindBrain(chan);

      if (!brain) return;

      len = brain->Chans.NumNodes();
      for(i=0;i<len;)
      {
         if (!strcmp(chan, brain->Chans.GetDatabyIndex(i)))
         {
            delete brain->Chans.GetDatabyIndex(i);
            brain->Chans.DeletebyIndex(i);
            len--;
         }
         else i++;
      }
   }
   else
   {
      len = Brains.NumNodes();
      brain = &Default;
      do
      {
         while(brain->Chans.NumNodes())
         {
            delete brain->Chans.GetDatabyIndex(0);
            brain->Chans.DeletebyIndex(0);
         }
         brain = Brains.GetDatabyIndex(i++);
      } while(i<=len);
   }
}


void SaveBrain(Brain *brain)
{
   char buf[300];

   printf(" Saving %s...\n", brain->FileName);
   wsprintf(buf, "%s.brn",brain->FileName);
   save_model(buf, brain->Model);
}


void LoadBrain(char *name)
{
   int i, len;
   Brain *brain;

   strupr(name);

   len = Brains.NumNodes();
   for(i=0;i<len;i++)
   {
      brain = Brains.GetDatabyIndex(0);
      if (!strcmp(brain->FileName, name)) return;
   }

   brain = new Brain;
   brain->dict = new_dictionary();
   change_personality(brain->dict, 0);
   load_model(name, &brain->Model, &brain->ban, &brain->aux, &brain->grt,
              &brain->swp, &brain->TrnFile);
   strcpy(brain->FileName, name);

   Brains.Add(brain);
}


void UnloadBrain(Brain *brain)
{
   if (brain->Log && brain->LogFile) fclose(brain->LogFile);
   if (brain->TrnFile) fclose(brain->TrnFile);
   Brains.DeletebyData(brain);
}

void Process(Brain *brain, char *str)
{
   char buf[2048];

   strcpy(buf, str);
   strupr(buf);

   SetBrain(brain);
   make_words(buf, Words);
}

void Learn(Brain *brain)
{
   SetBrain(brain);
   learn(brain->Model, Words);
}
   
int GetResponse(Brain *brain, char *str)
{
   char *temp;
   
   SetBrain(brain);

   temp = generate_reply(brain->Model, Words); 
   if (!temp || strlen(temp) == 0) return 1;
   capitalize(temp);
   strcpy(str, temp);
   return 0;
}


void SetBrain(Brain *brain)
{
   load_personality(brain->aux, brain->ban, brain->grt, brain->swp);
   ThinkTime   = brain->ThinkTime;
   SurpriseAvg = brain->SurpriseAvg;
}


Brain * ChooseBrain(bool showall)
{
   int i, len;
   char input[4];
   char c = 66, ci;

   puts(" Current Brains:\n"
        " ---------------");

   if (showall)
   {
      c = 67;
      puts(" A) All\n"
           " B) Default");
   }
   else
   {
      c = 66;
      puts(" A) Default");
   }

   len = Brains.NumNodes();
   for(i=0;i<len;i++) printf(" %c) %s\n", c++, Brains.GetDatabyIndex(i)->FileName);

   do
   {
      printf("\n Choose a brain: ");
      GetLine(input, 4);

      ci = toupper(input[0]);
   } while(ci < 65 || ci >= c);

   if (showall)
   {
      if (ci == 65) return NULL;
      if (ci == 66) return &Default;
      else return Brains.GetDatabyIndex(ci-67);
   }
   else
   {
      if (ci == 65) return &Default;
      else return Brains.GetDatabyIndex(ci-66);
   }
}


int TrainBrain(Brain *brain, char *fn)
{
   return train(brain->Model, fn);
}


void Ignore(Brain *brain, char *list, IgnCmd cmd)
{
   char buf[256];
   int i;

   while(*list)
   {
      i = 0;
      for(;*list && (isspace(*list)||*list == ',');list++);
      for(;*list && !isspace(*list) && *list != ',';buf[i++]=*list++);
      if (!i) return;
      buf[i] = 0;
      IgnoreCmd(brain, buf, cmd);
   }
}


void IgnoreCmd(Brain *brain, char *name, IgnCmd cmd)
{
   int i, num = brain->Ignores.NumNodes();
   char *temp;
   static char *MemErr = "Error Allocating Memory for IgnoreCmd.";

   strlwr(name);

   switch(cmd)
   {
      case IGN_ADD:
         if (*name == '#' || *name == '&')
         {
            for(i=num-1;i>=0;i--)
            {
               temp = brain->Ignores.GetDatabyIndex(i);
               if (*temp != '#' && *temp != '&') break;
               if(!strcmp(name, temp)) return;
            }
            temp = (char *)malloc(num=(strlen(name)+1));
            if (!temp)
            {
               puts(MemErr);
               return;
            }
            CopyMemory(temp, name, num);
            brain->Ignores.Add(temp);
         }
         else
         {
            for(i=0;i<num;i++)
            {
               temp = brain->Ignores.GetDatabyIndex(i);
               if (*temp == '#' || *temp == '&') break;
               if(!strcmp(name, temp)) return;
            }
            temp = (char *)malloc(num=(strlen(name)+1));
            if (!temp)
            {
               puts(MemErr);
               return;
            }
            CopyMemory(temp, name, num);
            brain->Ignores.Insert(temp, 0);
         }
         break;

      case IGN_REM:
         if (*name == '#' || *name == '&')
         {
            for(i=num-1;i>=0;i--)
            {
               temp = brain->Ignores.GetDatabyIndex(i);
               if (*temp != '#' && *temp != '&') break;
               if (!strcmp(name, temp))
               {
                  free(brain->Ignores.GetDatabyIndex(i));
                  brain->Ignores.DeletebyIndex(i);
                  return;
               }
            }
         }
         else
         {
            for(i=0;i<num;i++)
            {
               temp = brain->Ignores.GetDatabyIndex(i);
               if (*temp == '#' || *temp == '&') break;
               if (!strcmp(name, temp))
               {
                  free(brain->Ignores.GetDatabyIndex(i));
                  brain->Ignores.DeletebyIndex(i);
                  return;
               }
            }
         }
         break;

      case IGN_CLEAR:
         while(num--)
         {
            free(brain->Ignores.GetDatabyIndex(0));
            brain->Ignores.DeletebyIndex(0);
         }
         break;
   }
}


int Ignored(Brain *brain, char *name)
{
   int i, num = brain->Ignores.NumNodes();
   char *temp;

   if (!num) return 0;

   if (*name == '#' || *name == '&')
   {
      for(i=num-1;i>=0;i--)
      {
         temp = brain->Ignores.GetDatabyIndex(i);
         if (*temp != '#' && *temp != '&') break;
         if(!_stricmp(name, temp)) return 1;
      }
   }
   else
   {
      for(i=0;i<num;i++)
      {
         temp = brain->Ignores.GetDatabyIndex(i);
         if (*temp == '#' || *temp == '&') break;
         if(!_stricmp(name, temp)) return 1;
      }
   }

   return 0;
}


void LogBrain(Brain *brain, int log)
{
   char name[256];

   if (log)
   {
      if (brain->LogFile) return;
      brain->Log = 0;
      wsprintf(name, "%s.log", brain->FileName);
      brain->LogFile = fopen(name, "a+");
      if(!brain->LogFile) puts(" Error opening log file.");
      else brain->Log = 1;
   }
   else
   {
      brain->Log = 0;
      if (brain->LogFile) fclose(brain->LogFile);
      brain->LogFile = NULL;
   }
}


void AddLog(Brain *brain, char *ctrl, ...)
{
   char buf[1024];
   va_list  list;
   if (!brain) brain = &Default;
   int len;

   va_start(list, ctrl);

   len = vsprintf(buf, ctrl, list);
   buf[len] = '\n', buf[len+1]=0;

   if (brain->Log) fputs(buf, brain->LogFile);
   if (brain->VerbLog)
   {
      if (VerboseNicks[0]) PrintMessage(buf, VerboseNicks);
      buf[len] = 0;
      PrintMessage(buf);
   }

   va_end(list);
}
