/* By Adam Milazzo - 1999 */

#include <ctype.h>
#include <time.h>
#include "CIRC.h"


static CIRC *Me; // lame hack

#define IDENT "1826, 6667 : USERID : UNIX : "

CIRC::CIRC()
{
   CWCalls calls;

   OnMsg     = OnAction     = OnKick = NULL;
   OnConnect = OnDisconnect = NULL;
   OnVerbose = NULL;
   OnError   = NULL;
   Me        = this;
   Nick[0]   = 0;
   Verbose   = FALSE;

   InitializeCriticalSection(&CS);
   ZeroMemory(&calls, sizeof(CWCalls));
//   calls.Connected = OnConn;
   calls.Close     = OnClose;
   calls.Error     = OnErr;
   calls.WriteRdy  = OnConn;
   calls.Read      = OnRead;
   calls.ConnReq   = OnReq;

   strcpy(IdentStr, "huzzah");
   
   Socks.Init();
   Socks.SetCallbacks(&calls);
   Socks.OpenSocket(&MySock, CW_TCP, FALSE);

   Socks.OpenSocket(&Ident, CW_TCP, FALSE);
   Ident.Bind(113);
   Ident.Listen(1);
}


CIRC::~CIRC()
{
   Disconnect("");
   Socks.Deinit();
   DeleteCriticalSection(&CS);
}


void CIRC::SetVerbose(bool verbose)
{
   Verbose = verbose;
}


int CIRC::OnReq(CWSocket *cws)
{
   if (!Me->DoingIdent) Me->Ident.Accept(&Me->IdentConn);
   return 0;
}


int CIRC::OnRead(CWSocket *cws)
{
#define MAXLINE 1024
   static char Line[MAXLINE+1];
   static int  Length;
   int         ret;
   char       *pos, *temp;
   bool        discard=FALSE;

   if (Length == MAXLINE) // we're going to discard this line. it's too long
   {
      discard = TRUE;
      Length  = 0;
   }
   ret = cws->Recv(Line+Length, MAXLINE-Length, 0);
   if (ret == SOCKET_ERROR)
   {
      if (Me->OnError) Me->OnError(SOCKET_ERROR);
      return 1;
   }

   Line[Length+ret] = 0;

   temp = Line;
   while(pos = strchr(temp, '\n'))
   {
      *pos-- = 0;                     // DONT want LF
      if (*pos == '\r') *pos = 0;     // DONT want CR
      pos+=2;

      if (Me->Verbose && Me->OnVerbose) Me->OnVerbose(temp);
      if (!discard) Me->Parse(temp);
      temp = pos;
   }

   Length = Length + ret - (int)(temp - Line); // length to copy
   if (Length > 0) CopyMemory(Line, temp, Length);
   else Length = 0;

   return 0;
}


int CIRC::OnConn(CWSocket *cws)
{
   char buf[256];
   int i;

   if (cws == &Me->IdentConn)
   {
      char buf[64];
      wsprintf(buf, "%s%s \n", IDENT, Me->IdentStr);
      Me->DoingIdent = TRUE;
      cws->Send(buf, strlen(buf));
      cws->ShutDown();
      Me->DoingIdent = FALSE;
      return 0;
   }

   if (!Me->Nick[0])
   {
      srand(time(NULL)*(i|1)*(int)Me);
      for(i=0;i<8;i++) buf[i] = rand()%26 + 97;
      buf[8] = 0;
   }
   else strcpy(buf, Me->Nick);

   Me->Connected = TRUE;
   Me->SetNick(buf);
   wsprintf(buf, "USER %s B C :Hello :)\n", Me->IdentStr);
   Me->SendRaw(buf);

   if (Me->OnConnect) Me->OnConnect();
   return 0;
}


int CIRC::OnClose(CWSocket *cws)
{
   if (Me->OnDisconnect) Me->OnDisconnect();
   Me->Connected = FALSE;
   return 0;
}


int CIRC::OnErr(CWSocket *cws, int err)
{
//   char buf[128];

//   wsprintf(buf, "Winsock error: %d\n", err);
//   OutputDebugString(buf);

   Me->Socks.CloseSocket(cws, 0, 0);
   if (Me->OnError) Me->OnError(err);
   return 0;
}


char * CIRC::GetWord(char *line, char *word)
{
   int i,j,len;

   len = strlen(line);
   for(i=0;i<len;i++) if (!isspace(line[i])) break;
   for(j=0;i<len;i++,j++)
   {
      if (isspace(line[i])) break;
      word[j] = line[i];
   }
   word[j] = 0;

   return line+i;
}


char * CIRC::NextWord(char *line)
{
   while(*line)
   {
      if(!isspace(*line)) break;
      line++;
   }
   while(*line)
   {
      if(isspace(*line)) break;
      line++;
   }
   while(*line)
   {
      if(!isspace(*line)) break;
      line++;
   }
   return line;
}


void CIRC::Parse(char *line)
{
   char parm1[256], parm2[256], parm3[256], parm4[256], spec[1024+256];
   char *src, *dest, *text, *channel, *temp;
   int   i, j, len;

   // read the parameters
   GetWord(line, parm1);
   if (parm1[0] != ':')
   {
      strupr(parm1);
      if (strcmp(parm1, "PING")) return; // ignore it if it's not something we're interested in
   }

   line = NextWord(line);
   GetWord(line, parm2);
   line = NextWord(line);
   GetWord(line, parm3);

   if (temp = strchr(parm1, '!'))  // it's from a player
   {
      len = (int)(temp - parm1);
      for(i=1,j=0;i<len;i++,j++) parm1[j] = parm1[i];
      parm1[j] = 0;
      src = parm1; // src is the nick

      _strupr(parm2);

      if (OnMsg && !strcmp(parm2,"PRIVMSG"))
      {
         dest = parm3;
         if (dest[0] == '#' || dest[0] == '&') channel = parm3;
         else channel = NULL;
         line = NextWord(line);
         if (*line == ':') line++;
         GetWord(line, parm4);

         if (!strcmp(parm4, "\01ACTION"))
         {
            return; // SKIPPING ACTIONS!, remove this if you want it to learn from actions
            line = NextWord(line); // skip the action
            strcpy(spec, src);
            strcat(spec, " ");
            strcat(spec, line);
            text = spec;
         }
         else if(parm4[0] == 1) return; // other commands, ctcp, dcc, etc
         else text = line;

         len = strlen(text);
         if (!len) return;

         if (text[len-1] == '\01') text[len-1] = 0;
         
         Strip(text);
         OnMsg(src, dest, channel, text);
      }
      else if(OnKick && !strcmp(parm2, "KICK"))
      {
         line = NextWord(line);
         GetWord(line, parm4);
         line = NextWord(line);

         if (*line == ':') line++;

         channel = parm3;
         dest = parm4;
         text = line;

         Strip(text);
         OnKick(src, dest, channel, text);
      }
      else return;
   }
   else // from the server
   {
      if (!strcmp(parm1, "PING"))
      {
         char buf[256];
         wsprintf(buf, "PONG %s\n", parm2);
         SendRaw(buf);
      }
      else return;
   }

   return;
}


void CIRC::Strip(char *line)
{
   int i, j, len;
   char c;

   len = strlen(line);
   for(i=j=0;i<len;i++)
   {
      c = line[i];
      if (c >= 32 && c <= 126) line[j++] = c;
      else if (c == 3)
      {
         if (isdigit(line[i+1])) i++;
         if (isdigit(line[i+1])) i++;
      }
   }
   line[j] = 0;
}


int CIRC::Connect(char *server, int port)
{
   int ret;
   
   Disconnect("");
   ret = MySock.Connect(server, port);
   if ((ret == SOCKET_ERROR && WSAGetLastError() == WSAEWOULDBLOCK) || ret == 0) return 0;
   return 1;
}


void CIRC::Disconnect(char *reason)
{
   char buf[300];

   if (Connected)
   {
      wsprintf(buf, "QUIT :%s\n", reason);
      SendRaw(buf);
      Sleep(500);
   }

   Connected = FALSE;
}


void CIRC::SetIdent(char *str)
{
   strcpy(IdentStr, str);
}


char * CIRC::GetIdent()
{
   return IdentStr;
}


char * CIRC::GetNick()
{
   return Nick;
}


void CIRC::SetNick(char *nick)
{
   char buf[300];

   strcpy(Nick, nick);
   wsprintf(buf, "NICK %s\n", nick);
   if (Connected) SendRaw(buf);
}


void CIRC::SendMessage(char *dest, char *msg)
{
   char buf[1100];

   wsprintf(buf, "PRIVMSG %s :%s\n", dest, msg);
   SendRaw(buf);
}


void CIRC::SendRaw(char *data)
{
   EnterCriticalSection(&CS);
   MySock.Send(data, strlen(data));
   LeaveCriticalSection(&CS);
}


void CIRC::Join(char *channel)
{
   char buf[300];

   wsprintf(buf, "JOIN %s\n", channel);
   SendRaw(buf);
}


void CIRC::Part(char *channel)
{
   char buf[300];

   wsprintf(buf, "PART %s\n", channel);
   SendRaw(buf);
}
