/* CWinsock - Multithreaded version */
#ifndef _CWINSOCK_H
#define _CWINSOCK_H

#include <winsock2.h> // must go BEFORE windows.h
#include <windows.h>
#include <vector>

struct CWSockOpts
{
   int debug;        // set if you want winsock to output debug info
   int dontlinger;   // clear if you want it to linger on socket closure
   int dontroute;    // dont touch??
   int keepalive;    // set if you want TCP connection to be tested by keepalives
   struct linger lingerset; // linger settings; used if dontlinger is cleared
   int oobinline;    // set if you want OOB data to be receieved like other dadta
   int reuseaddr;    // set to allow binding to an already bound port
   int nodelay;      // set to disable Nagle algorithm for TCP.. not recommended. [Nagle algorithm combines many small packets into larger ones]
   int rcvbuf;       // set the size of the receive buffer. must re-read this to check the actual size obtained.
   int sndbuf;       // set the size of the send buffer. must re-read this to check the actual size obtained.
   u_long maxsize;   // read only, max UDP message size
   u_long nonblocking; // dont use this one :) it's for the class' use only
};


enum CWType
{
   CW_TCP = SOCK_STREAM, CW_UDP = SOCK_DGRAM
};

#define CW_PEEK MSG_PEEK

enum CWError
{
   CWE_READ = 0x01, CWE_WRITE = 0x02, CWE_ACCEPT = 0x04, CWE_CONNECT = 0x08, CWE_CLOSE = 0x10
};


class CWIPAddress
{
   private:
      sockaddr_in Address;

   public:
      CWIPAddress();       // empty IP
      CWIPAddress(char *, int); // from ip string and port
      CWIPAddress(DWORD, int);  // from a long IP address and port
      CWIPAddress(sockaddr_in *); // from a sockaddr_in *

      int   SetIPAddress(char *, int);      // set the ip address
      int   SetIPAddress(DWORD, int);
      int   SetIPAddress(sockaddr_in *);

      char * GetIPString();   // returns * to a static buffer.. dont modify.. copy if needed
      DWORD  GetIPLong();
      sockaddr_in * GetIPSockAddr(); // returns pointer to actual Address.. bad style i know.. so sue me..
};


class CWinsock;
struct CWSocket
{
      CWSocket() // inline constructor
      {
         hEvent = NULL;
         hSock  = INVALID_SOCKET;
         Parent = NULL;
         myOpts.lingerset.l_onoff = 0;
      };
      int         Bind(int port);
      int         Bind(char *addr, int port); // for multiple IPs

      int         Listen(int log); // num conns to queue, 0 for default
      
      int         Accept(CWSocket *newsock); // accept an incoming and pass onto newsock
      
      int         Connect(char *addr, int port);

      int         ShutDown();  // shutdown a socket (disallow sends and cause FD_CLOSE
                               // event to occur on other side)
      // for UDP sockets, these send to and retrieve from the default address
      int         Send(char *buf, int len);
      int         Recv(char *buf, int len, int flags=0); // flags can be CW_PEEK

      // these send to an ip address and retrieve data and store the ip address it was from
      // designed for message-based (UDP) sockets only
      int         SendTo(char *buf, int len, CWIPAddress *addr);  // for UDP
      int         RecvFrom(char *buf, int len, int flags, CWIPAddress *addr); // UDP

      int         SetOpts(CWSockOpts *);
      int         GetOpts(CWSockOpts *);

      SOCKET      GetSocket() const;

      // returns amount of data that can be read in a single Recv call (may be less than actual data queued)
      int         DataWaiting() const;

      // return IP of connected peer or NULL on error
      char const * GetPeerIP() const;
      // return port of connected peer or -1 on error
      int         GetPeerPort() const;

      CWSockOpts  myOpts;
      HANDLE      hEvent;
   
      SOCKET      hSock;
      CWType      Type;
      char        RemAddr[16];
      int         RemPort;

      CWinsock   *Parent;
};


struct CWCalls
{
   int (* Read)      (CWSocket *);
   int (* WriteRdy)  (CWSocket *);
   int (* ConnReq)   (CWSocket *);
   int (* Connected) (CWSocket *);
   int (* Close)     (CWSocket *);
   int (* Error)     (CWSocket *, int);
};


class CWinsock
{
   protected:
      HANDLE *Events;
      int     nEvents, maxEvents;

      HANDLE                  hRebuild, hKill;
      HANDLE                  hThread;
      CRITICAL_SECTION        CW_CS;
      std::vector<CWSocket *> Socks;
      CWCalls                 CallBacks;
      WSADATA                 WSData;
      volatile int            sThread, sRebuild;
      bool                    bInit;

      int         InitThread();
      void        KillThread();
      void        RebuildEvents();

      int         Open(CWSocket *, CWType, bool blocking);
      int         Close(CWSocket *); // close, use default linger settings
      // if linger, wait timeout seconds or until all data has been sent
      int         Close(CWSocket *, BOOL lingon, int timeout); 

      static int DefRead     (CWSocket *);
      static int DefWriteRdy (CWSocket *);
      static int DefConnReq  (CWSocket *);
      static int DefConnected(CWSocket *);
      static int DefClose    (CWSocket *);
      static int DefError    (CWSocket *, int);

      static DWORD WINAPI ThreadFunc(LPVOID);

   public:
      CWinsock();
      ~CWinsock();

      int      Init();
      void     Deinit();
      
      void     SetCallbacks(CWCalls *);
      void     GetCallbacks(CWCalls *) const;

      int      OpenSocket(CWSocket *); // used internally... not too sure why it's public :)
      int      OpenSocket(CWSocket *, CWType type, bool blocking);
      int      OpenSocket(CWSocket *, SOCKET s, CWType type, bool blocking); // use existing socket

      int      CloseSocket(CWSocket *); // uses default linger settings
      // if linger, wait timeout seconds or until all data has been sent
      int      CloseSocket(CWSocket *, bool lingon, int timeout);

      // checks for socket events and processes them if necessary
      inline void PollSockets() {}; // does nothing in multithreaded version

      int      GetLocalInfo(char *addr, int len) const; // name of local machine

      int      LastError() const; // uses winsock error code

      int      GetWSData(WSADATA *) const;

      friend   DWORD WINAPI ThreadFunc(LPVOID);
      friend   class CWIPAddress;
};

#endif
