/* By Adam Milazzo - 1999 */

#ifndef _CLINKEDLIST_H
#define _CLINKEDLIST_H

#include <windows.h>

template <class Type>
class CLinkedList
{
   protected:
      struct CLinkedListNode;
      struct CLinkedListNode
      {
         CLinkedListNode *Next;
         CLinkedListNode *Prev;
         Type             Data;
      };

      CLinkedListNode *First, *Last;
      DWORD nNodes;

      void * GetNodebyIndex(int);

   public:

      CLinkedList();
      ~CLinkedList();

      void                 Add(Type);           // add to the end
      int                  Insert(Type, DWORD); // insert at index
      void                 Clear();             // delete all nodes
      void                 DeletebyData(Type);  // delete the first match of this data
      void                 DeletebyIndex(int);  // delete this
      Type                 GetDatabyIndex(int); // get data from this index
      int                  FindbyData(Type);    // get an index from this value

      int                  NumNodes();          // returns the number of nodes
};

template <class Type> CLinkedList<Type>::CLinkedList()
{
   First = Last = NULL;
   nNodes = 0;
} /* CLinkedList */


template <class Type> CLinkedList<Type>::~CLinkedList()
{
   Clear();
} /* ~CLinkedList */


template <class Type> void * CLinkedList<Type>::GetNodebyIndex(int index)
{
   CLinkedListNode *temp = First;
   int i;

   if (index > (long)nNodes || index < 0) return NULL;

   for(i=0;i<index;i++) temp = temp->Next;

   return temp;
}

template <class Type> void CLinkedList<Type>::Add(Type data)
{
   CLinkedListNode   *temp;

   nNodes++;

   if (!First) // empty list
   {
      First = new CLinkedListNode;
      First->Next = First->Prev = NULL;
      First->Data = data;
      Last = First;
      return;
   }

   if (First == Last) // only one node
   {
      temp = new CLinkedListNode;
      temp->Prev = First;
      temp->Next = NULL;
      temp->Data = data;
      First->Next = temp;
      Last = temp;
   }
   else
   {
      temp = new CLinkedListNode;
      temp->Prev = Last;
      temp->Next = NULL;
      temp->Data = data;
      Last->Next = temp;
      Last = temp;
   }
} /* Add */


template <class Type> int CLinkedList<Type>::Insert(Type data, DWORD index)
{
   CLinkedListNode   *temp, *node;

   if (index == 0) // inserting at the beginning
   {
      temp = new CLinkedListNode;
      temp->Next = First;
      temp->Prev = NULL;
      temp->Data = data;
      if (First)
      {
         First->Prev = temp;
         First = temp;
      }
      else First = Last = temp;
         
   }
   else if (index == nNodes)
   {
      Add(data);
      return 0;
   }
   else
   {
      node = (CLinkedListNode *)GetNodebyIndex(index);
      if (!node) return 1;

      temp = new CLinkedListNode;
      temp->Next = node;
      temp->Prev = node->Prev;
      temp->Data = data;
      node->Prev->Next = temp;
      node->Prev = temp;
   }
   nNodes++;

   return 0;
} /* Insert */


template <class Type> void CLinkedList<Type>::Clear()
{
   while(First) DeletebyIndex(0);
}


template <class Type> void CLinkedList<Type>::DeletebyData(Type data)
{
   int i = FindbyData(data);
   
   if (i != -1) DeletebyIndex(i);
}


template <class Type> void CLinkedList<Type>::DeletebyIndex(int index)
{
   CLinkedListNode *dead = (CLinkedListNode *)GetNodebyIndex(index);

   if (dead == NULL) return;
   
   nNodes--;

   if (dead == First)
   {
      if (dead->Next)
      {
         dead->Next->Prev = NULL;
         First = dead->Next;
      }
      else First = NULL;
   }
   else if (dead == Last)
   {
      dead->Prev->Next = NULL;
      Last = dead->Prev;
   }
   else
   {
      dead->Prev->Next = dead->Next;
      dead->Next->Prev = dead->Prev;
   }

   delete dead;
}


template <class Type> Type CLinkedList<Type>::GetDatabyIndex(int index)
{
   CLinkedListNode *node = (CLinkedListNode *)GetNodebyIndex(index);

   if (node == NULL) return NULL;

   return node->Data;
}


template <class Type> int CLinkedList<Type>::FindbyData(Type data)
{
   CLinkedListNode *temp = First;
   int i=0;

   if (temp == NULL) return -1;

   do
   {
      if (temp->Data == data) return i;
      temp = temp->Next;
      i++;
   } while(temp);
   
   return -1;
}


template <class Type> int CLinkedList<Type>::NumNodes()
{
   return nNodes;
}

#endif