/* By Adam Milazzo - 1999 */

#include "CIRC.h"
#include <windows.h>
#include <stdio.h>
#include <conio.h>
#include <signal.h>
#include "megahal.h"
#include "Brains.h"
#include "MsgQueue.h"

#define VERSION     "2.33b"
#define DEFAULTPORT 6667
// ThinkTime
// seconds allowed to think for each response..
// higher means a better response, but a slower bot :)
// default for MegaHAL = 2000

// SurpriseAvg
// by default, megahal strives to find the most surprising reponse.
// HalBot tries to find a somewhat surprising response
// setting this higher raises the average surprise factor of the replies
// setting it to 100 will make megahal find the most surprising..
// actual max surprise is 16.0 or 20.0 or so.. anothing over the max will
// cause megahal to find the most surprising response

static CRITICAL_SECTION CS;

static volatile bool Connected, ExpectClose, TryReconnect, DoQuit, OkQuit = TRUE;

static char Nick[64];
static char Server[256];
static int  Port;

static int  Frequency = 40; // randomly speak one in 40 times (from channel messages)
static bool OnlyMsg=false,AutoReconnect=true,Verbose=false,LearnUnattached=true, IgnoreInChan=false;
// 200 ticks per char... 200/1000 = 5cps
static int  SpeechDelay = 200;

static void OnMsg   (char *src,    char *dest,   char *channel, char *text);
static void OnAction(char *src,    char *dest,   char *channel, char *text);
static void OnKick  (char *kicker, char *kicked, char *channel, char *text);
static void OnVerbose(char *msg);
static void OnConnect();
static void OnDisconnect();
static void OnError(int);

static char * StripNick(char *str);
static void HandleInput();
static void TrimSpace(char *str);
void GetLine(char *line, int len);
void DoCommaSep(char *str, void (* Funct)(char *));

static void AddChan(char *chan);
static void RemChan(char *chan);

extern CLinkedList<Brain *> Brains;
CLinkedList<char *>         Chans;

CIRC           IRC;
Brain          Default;
DICTIONARY    *Words;
volatile bool  HandlingInput;
char VerboseNicks[256];

void HandleInput()
{
   char input[1024];

   if (!kbhit()) return;

   HandlingInput = TRUE;

   printf("HB> ");
   GetLine(input, 1024);

   switch(toupper(input[0]))
   {
      case '?':
         printf("E  Enter new ident              (%s)\n"
                "N  Change nick                  (%s)\n"
                "C  Connect to server\n"
                "J  Join channel(s)\n"
                "P  Part channel(s)\n"
                "L  Load Brainfile\n"
                "U  Unload Brainfile\n"
                "S  Save brainfile\n"
                "SA Set Surprise Average\n"
                "A  Attach brain\n"
                "D  Detach brain\n"
                "B  View brain\n"
                "T  Train brain\n"
                "TT Think Time per message\n"
                "LN Toggle Brain Learning\n"
                "LU Unattached chans use Default (%d)\n"
                "K  Mute Brain\n"
                "G  Generate Response\n"
                "Z  Verbose Mode                 (%d)\n"
                "H  Human Factor [delay]         (%d)\n"
                "F  Set Speech Frequency         (%d)\n"
                "O  Only speak in privmsg        (%d)\n"
                "Y  Automatically Reconnect      (%d)\n"
                "[press a key for next page]",
                IRC.GetIdent(), IRC.GetNick(), LearnUnattached, Verbose, 
                SpeechDelay, Frequency, OnlyMsg, AutoReconnect);
                getch();
                putchar('\n');
         printf("W  Toggle Logging\n"
                "WT Write training file?\n"
                "V  Verbose Logging\n"
                "VI Verbose Logging to IRC       (%s)\n"
                "I  Insert String (manual train)\n"
                "IA Add user/chan to ignore list\n"
                "IR Remove user/chan from ignores\n"
                "IC Clear ignore list\n"
                "II Ignore users in a channel?   (%d)\n"
                "M  Send message\n"
                "MD Send delayed message\n"
                "MC Clear message queue\n"
                "MF Flush message queue\n"
                "R  Send raw command\n"
                "Q  Disconnect from IRC\n"
                "X  Exit HalBot\n",
                VerboseNicks, IgnoreInChan);
         break;

      case 'E':
         if (input[1]) IRC.SetIdent(input+2);
         else
         {
            printf(" Enter new ident: ");
            GetLine(input, 32);
            IRC.SetIdent(input);
         }
         break;

      case 'N':
         if (input[1]) strcpy(Nick, input+2);
         else
         {
            printf(" Nick: ");
            GetLine(input, 64);
            strcpy(Nick, input);
         }
         IRC.SetNick(Nick);
         strlwr(Nick);
         break;

      case 'C':
         {
            int port=0;
            TryReconnect = FALSE;
            if (input[1]) sscanf(input+2,"%s %d", input, &port);
            else
            {
               printf(" Enter <server> <port>: ");
               scanf("%s %d", input, &port);
            }
            if(!port) port=DEFAULTPORT;
            ExpectClose = FALSE;
            strcpy(Server,input);
            Port = port;
            puts(" Attempting Connection...");
            IRC.Disconnect("");
            IRC.Connect(input, port);
         }
         break;
      
      case 'J':
         if (input[1])
         {
            IRC.Join(input+2);
            DoCommaSep(input+2, AddChan);
         }
         else
         {
            printf(" Channel(s): ");
            GetLine(input, 1024);
            IRC.Join(input);
            DoCommaSep(input, AddChan);
         }
         break;

      case 'P':
         if (input[1])
         {
            IRC.Part(input+2);
            DoCommaSep(input+2, RemChan);
         }
         else
         {
            printf(" Channel(s): ");
            GetLine(input, 1024);
            IRC.Part(input);
            DoCommaSep(input, RemChan);
         }
         break;

      case 'L':
         if (toupper(input[1]) == 'N')
         {
            int canlearn;
            Brain *brain = ChooseBrain(TRUE);

            if (input[2]) sscanf(input+3, "%d", &canlearn);
            else
            {
               printf(" Can I learn? (0/1): ");
               scanf("%d", &canlearn);
            }
            canlearn = (canlearn != 0);
            if (!brain)
            {
               ALLBRAINS(brain->CanLearn = canlearn);
            }
            else brain->CanLearn = canlearn;
         }
         else if (toupper(input[1]) == 'U')
         {
            if (input[2]) sscanf(input+3, "%d", &LearnUnattached);
            else
            {
               printf(" Default learns from unattached chans? (0/1): ");
               scanf("%d", &LearnUnattached);
            }
            LearnUnattached = (LearnUnattached != 0);
         }
         else
         {
            if (input[1]) LoadBrain(input+2);
            else
            {
               printf(" Brainfile: ");
               GetLine(input, 256);
               LoadBrain(input);
            }
         }
         break;

      case 'U':
         {
            EnterCriticalSection(&CS);
            Brain *brain = ChooseBrain(TRUE);
            if (!brain)
            {
               ALLBRAINS(UnloadBrain(brain));
            }
            else UnloadBrain(brain);
            LeaveCriticalSection(&CS);
         }
         break;

      case 'S':
         if (toupper(input[1]) == 'A')
         {
            Brain *brain = ChooseBrain(TRUE);
            float surprise;
            if (input[2]) sscanf(input+3, "%f", &surprise);
            else
            {
               printf(" Enter new surprise average (0.0 - 100.0): ");
               scanf("%f", &surprise);
            }
            if (brain) brain->SurpriseAvg = surprise;
            else ALLBRAINS(brain->SurpriseAvg = surprise);
         }
         else
         {
            Brain *brain = ChooseBrain(TRUE);
            EnterCriticalSection(&CS);
            if (!brain)
            {
               ALLBRAINS(SaveBrain(brain));
            }
            else SaveBrain(brain);
            LeaveCriticalSection(&CS);
         }
         break;
               
      case 'A':
         {
            Brain *brain = ChooseBrain(FALSE);
            if (input[1]) Attach(brain, input+2);
            else
            {
               printf(" Channel: ");
               GetLine(input, 1024);
               Attach(brain, input);
            }
         }
         break;

      case 'D':
         if (input[1]) Detach(input+2);
         else
         {
            printf(" Channel: ");
            GetLine(input, 1024);
            Detach(input);
         }
         break;

      case 'B':
         {
            Brain *brain = ChooseBrain(FALSE);
            ShowBrain(brain);
         }
         break;

      case 'T':
         if (toupper(input[1]) == 'T')
         {
            Brain *brain = ChooseBrain(TRUE);
            int thinktime;
            if (input[2]) sscanf(input+3, "%d", &thinktime);
            else
            {
               printf(" Enter thinking time allowed in milliseconds: ");
               scanf("%d", &thinktime);
            }
            if (brain) brain->ThinkTime = thinktime;
            else ALLBRAINS(brain->ThinkTime = thinktime);
         }
         else
         {
            Brain *brain = ChooseBrain(FALSE);
            char *str = input;
            int ret;
            if (input[1]) ret = TrainBrain(brain, str+=2);
            else
            {
               printf(" TextFile: ");
               GetLine(input, 1024);
               EnterCriticalSection(&CS);
               ret = TrainBrain(brain, input);
               LeaveCriticalSection(&CS);
            }
            if (!ret && brain->WriteTrn && brain->TrnFile)
            {
               char resp[10];
               printf(" Add to train file? (Y/N) ");
               GetLine(resp, 10);
               if(toupper(resp[0]) == 'Y')
               {
                  FILE *trn = fopen(str, "r");

                  if (!trn)
                  {
                     printf("  Failed to open file %s.\n", str);
                     break;
                  }
                  while(!feof(trn))
                  {
                  	if(fgets(input, 1024, trn)==NULL) break;
                     TrimSpace(input);
                     fprintf(brain->TrnFile, "%s\n", input);
                  }
                  fclose(trn);
               }
            }
         }
         break;

      case 'Z':
         if (input[1]) sscanf(input+2, "%d", &Verbose);
         else
         {
            printf(" Verbose? (0/1): ");
            scanf("%d", &Verbose);
         }
         Verbose = (Verbose != 0);
         IRC.SetVerbose(Verbose);
         break;

      case 'H':
         if (input[1]) sscanf(input+2, "%d", &SpeechDelay);
         else
         {
            printf(" Enter Delay (x ms/char): ");
            scanf("%d", &SpeechDelay);
         }
         break;

      case 'F':
         if (input[1]) sscanf(input+2, "%d", &Frequency);
         else
         {
            printf(" Enter Frequency: ");
            scanf("%d", &Frequency);
         }
         break;

      case 'O':
         if (input[1]) sscanf(input+2, "%d", &OnlyMsg);
         else
         {
            printf(" Only respond to privmsg? (0/1): ");
            scanf("%d", &OnlyMsg);
         }
         OnlyMsg = (OnlyMsg != 0);
         break;

      case 'G':
         {
            char buf[2048],*in=input;
            Brain *brain = ChooseBrain(FALSE);
            if (input[1]) in += 2;
            else
            {
               printf("Enter sentence: ");
               GetLine(input, 1024);
            }

            EnterCriticalSection(&CS);
            Process(brain, in);
            if (brain->CanLearn) Learn(brain);
            if (GetResponse(brain, buf)) strcpy(buf, "Error: No response.");
            LeaveCriticalSection(&CS);
            puts(buf);
         }
         break;

      case 'K':
         {
            int ismute;
            Brain *brain = ChooseBrain(TRUE);
            if (input[1]) sscanf(input+2, "%d", &ismute);
            else
            {
               printf(" Mute Brain? (0/1): ");
               scanf("%d", &ismute);
            }
            ismute = (ismute != 0);
            if (!brain)
            {
               ALLBRAINS(brain->IsMute = ismute);
            }
            else brain->IsMute = ismute;
            break;
         }

      case 'Y':
         if (input[1]) sscanf(input+2, "%d", &AutoReconnect);
         else
         {
            printf(" Automatically Reconnect? (0/1): ");
            scanf("%d", &AutoReconnect);
         }
         AutoReconnect = (AutoReconnect != 0);
         break;

      case 'W':
         if (toupper(input[1]) == 'T')
         {
            int writetrn;
            Brain *brain = ChooseBrain(TRUE);

            if (input[2]) sscanf(input+3, "%d", &writetrn);
            else
            {
               printf(" Write training file? (0/1): ");
               scanf("%d", &writetrn);
            }
            writetrn = (writetrn != 0);
            if (!brain)
            {
               ALLBRAINS(brain->WriteTrn = writetrn);
            }
            else brain->WriteTrn = writetrn;
         }
         else
         {
            int writelog;
            Brain *brain = ChooseBrain(TRUE);

            if (input[2]) sscanf(input+2, "%d", &writelog);
            else
            {
               printf(" Write log file? (0/1): ");
               scanf("%d", &writelog);
            }
            if (brain) LogBrain(brain, writelog);
            else ALLBRAINS(LogBrain(brain, writelog));
         }
         break;

      case 'V':
         {
            if(toupper(input[1]) == 'I')
            {
               if(input[2]) strcpy(VerboseNicks, input+3);
               else
               {
                  printf("Log to Nicks/Chans (Enter for none): ");
                  GetLine(VerboseNicks, 256);
               }
            }
            else
            {
               Brain *brain = ChooseBrain(TRUE);
               int verblog;
               if (input[1]) sscanf(input+2, "%d", &verblog);
               else
               {
                  printf(" Verbose Logging? (0/1): ");
                  scanf("%d", &verblog);
               }
               verblog = (verblog != 0);
               if (!brain)
               {
                  ALLBRAINS(brain->VerbLog = verblog);
               }
               else brain->VerbLog = verblog;
            }
         }
         break;

      case 'I':
         {
            Brain *brain;
            char *str = input;
            char cmd = toupper(input[1]);

            if(cmd == 'I')
            {
               if(input[2]) sscanf(input+3, "%d", &IgnoreInChan);
               else
               {
                  printf(" Ignore ignored users even in channel? (0/1): ");
                  scanf("%d", &IgnoreInChan);
               }
               IgnoreInChan = (IgnoreInChan != 0);

            }

            if(cmd == 'A' || cmd == 'R' || cmd == 'C') brain = ChooseBrain(TRUE);
            else brain = ChooseBrain(FALSE);

            if(cmd == 'A')
            {
               if (input[2]) str += 3;
               else
               {
                  printf(" Ignore what users/chans? ");
                  GetLine(input, 1024);
               }
               if (brain) Ignore(brain, str, IGN_ADD);
               else ALLBRAINS(Ignore(brain, str, IGN_ADD))
            }
            else if(cmd == 'R')
            {
               if (input[2]) str += 3;
               else
               {
                  printf(" Unignore what users/chans? ");
                  GetLine(input, 1024);
               }
               if (brain) Ignore(brain, str, IGN_REM);
               else ALLBRAINS(Ignore(brain, str, IGN_REM))
            }
            else if(cmd == 'C')
            {
               if (brain) Ignore(brain, str, IGN_CLEAR);
               else ALLBRAINS(Ignore(brain, str, IGN_CLEAR))
            }
            else
            {
               if (input[1]) str += 2;
               else
               {
                  printf(" Sentence: ");
                  GetLine(input, 1024);
               }
               EnterCriticalSection(&CS);
               Process(brain, str);
               Learn(brain);
               if (brain->WriteTrn && brain->TrnFile) fprintf(brain->TrnFile, "%s\n", str);
               LeaveCriticalSection(&CS);
            }
         }
         break;
         
      case 'M':
         if (toupper(input[1]) == 'C')
         {
            puts(" Clearing send queue.");
            ClearMessageQueue();
         }
         else if (toupper(input[1]) == 'F')
         {
            puts(" Flushing send queue.");
            FlushMessageQueue();
         }
         else
         {
            char dest[256];
            int i=0,j,len, delayed=0;
            if (toupper(input[1]) == 'D') delayed=i=2; // i = 2, delayed = TRUE
            else i = 1;
            if (input[1])
            {
               len=strlen(input);
               for(   ;i<len;i++) if (!isspace(input[i])) break;
               for(j=0;i<len && j<255;i++)
               {
                  if (!isspace(input[i])) dest[j++] = input[i];
                  else break;
               }
               dest[j] = 0;
               for(   ;i<len;i++) if (!isspace(input[i])) break;
            }
            else
            {
               printf(" Nick/Chan: ");
               GetLine(dest, 256);
               printf(" Message: ");
               GetLine(input, 1024);
            }
            AddLog(NULL, "To %s: %s",dest, input+i);
            if (!delayed) IRC.SendMessage(dest, input+i);
            else StartMessage(dest, input+i, strlen(input+i) * SpeechDelay);
         }
         break;

      case 'R':
         if (input[1])
         {
            int len = strlen(input);
            input[len] = '\n';
            input[len+1] = 0;
            IRC.SendRaw(input+2);
         }
         else
         {
            printf(" Command: ");
            fgets(input, 1024, stdin);
            IRC.SendRaw(input);
         }
         break;
         
      case 'Q':
         {
            char resp[16];

            StopMessageQueue();
            ExpectClose = TRUE;
            TryReconnect = FALSE;
            printf(" Quit: Are you sure? (Y/N): ");
            GetLine(resp, 16);
            if (toupper(resp[0]) != 'Y')
            {
               puts(" Aborted.");
               break;
            }
            if (input[1]) IRC.Disconnect(input+2);
            else
            {
               printf(" Reason: ");
               GetLine(input, 1024);
               IRC.Disconnect(input);
            }
            break;
         }

      case 'X':
         {
            bool valid=FALSE,quit;
            char resp[16];

            while(!valid)
            {
               printf(" Exit: Save All Brains? (Y/N/C): ");
               GetLine(resp, 16);
               valid = TRUE; // assume valid response
               switch(toupper(resp[0]))
               {
                  case 'Y':
                     EnterCriticalSection(&CS);
                     ALLBRAINS(SaveBrain(brain));
                     LeaveCriticalSection(&CS);
                     quit = TRUE;
                     break;

                  case 'N':
                     quit = TRUE;
                     break;

                  case 'C':
                     quit = FALSE;
                     puts(" Aborted.");
                     break;

                  default:
                     valid = FALSE;
                     break;
               }
            }

            if (quit)
            {
               StopMessageQueue();
               ExpectClose = TRUE;
               TryReconnect = FALSE;
               DoQuit = TRUE;
               IRC.Disconnect("");
            }
         }
         break;
   }

   HandlingInput = FALSE;
}


void OnMsg(char *src,    char *dest,   char *channel, char *text)
{
   char *to,str[2048];
   Brain *brain;
   bool respond=0, name=0;
   DWORD curtime = GetTickCount();

   if (!_stricmp(Nick, src)) return; // dont respond to ourselves!

   if (channel) brain = FindBrain(channel);
   else brain = &Default;
   if (!brain) if (LearnUnattached) brain = &Default; else return;

   if (channel) strlwr(channel);

   if(channel)
   {
      if (IgnoreInChan) if (Ignored(brain, src)) return;
      if (Ignored(brain, channel)) return;
   }
   else if (Ignored(brain, src)) return;

   TrimSpace(text);
   strcpy(str, text);
   strlwr(str);
   
   if (!brain->IsMute)
   {
      if (!_stricmp(dest, Nick))  // respond to a privmsg
      {
         to = src;
         respond = 1;

         AddLog(brain, "<%s> %s", to, text);
      }
      else if (!OnlyMsg && strstr(str, Nick)) // somebody says my name in a channel
      {
         to = channel;
         respond = 1;
         name = 1;

         AddLog(brain, "%s:<%s> %s", channel, src, text);
      }
      else if (!OnlyMsg)
      {
         int i = rand();
         if (Frequency && rand() % Frequency == 0) // respond randomly
         {
            to = channel;
            respond = 1;
            name = 1;

            AddLog(brain, "%s:<%s> %s", channel, src, text);
         }
      }
   }

   StripNick(str);
   if (DoQuit) return;
   OkQuit = FALSE;
   EnterCriticalSection(&CS);
//OutputDebugString("OnMsg - Enter\n");
   Process(brain, str);
   if (brain->CanLearn) Learn(brain);
   if (brain->WriteTrn && brain->TrnFile) fprintf(brain->TrnFile, "%s\n", str);
   if (brain->IsMute) goto End;

   str[0] = 0;

   if (respond)
   {
      if (name && rand() & 1)
      {
         strcpy(str, src);
         strcat(str, ": ");
      }

      if (to == src) strlwr(src);

      if (GetResponse(brain, str+strlen(str))) goto End; // couldnt generate output

      AddLog(brain, "%s",str);

      StartMessage(to, str, curtime - GetTickCount() + SpeechDelay*strlen(str));
   }
   
   End:
//OutputDebugString("OnMsg - Leave\n");
   LeaveCriticalSection(&CS);
   OkQuit = TRUE;
}


int main()
{
   InitializeCriticalSection(&CS);
   IRC.OnMsg        = OnMsg;
   IRC.OnAction     = OnMsg;
   IRC.OnKick       = OnKick;
   IRC.OnConnect    = OnConnect;
   IRC.OnDisconnect = OnDisconnect;
   IRC.OnError      = OnError;
   IRC.OnVerbose    = OnVerbose;

   if (InitMessageQueue())
   {
      puts("Error initializing message queue.");
      return 1;
   }
   signal(SIGINT, SIG_IGN);
   //srand(GetTickCount());
   Server[0] = 0;
   puts("HalBot " VERSION " Initialized.");

   Words = new_dictionary();
   Default.dict = new_dictionary();
   change_personality(Default.dict, 0);
   load_model("Default", &Default.Model, &Default.ban, &Default.aux, &Default.grt,
              &Default.swp, &Default.TrnFile);
   strcpy(Default.FileName, "Default");
   
   putchar('\n');

   while(!DoQuit)
   {
      if (TryReconnect)
      {
         int i;
         putchar('\n');

         while(!Connected)
         {
            puts("Attempting Reconnection...");
            Sleep(1000);
            IRC.Connect(Server, Port);
            for(i=0;i<25;i++)
            {
               Sleep(200);
               if(Connected) break;
               HandleInput();
            }

            if (!AutoReconnect)
            {
               TryReconnect = FALSE;
               break;
            }
         }

         Sleep(2000);
         int num = Chans.NumNodes();
         for(i=0;i<num;i++)
         {
            IRC.Join(Chans.GetDatabyIndex(i));
            Sleep(100);
         }
         TryReconnect = FALSE;
      }
      HandleInput();
      Sleep(0);
   }

   while(!OkQuit) Sleep(0);
   DeinitMessageQueue();
   DeleteCriticalSection(&CS);
   puts("\nHalBot Shutdown.\n");
   return 0;
}


void OnKick(char *kicker, char *kicked, char *channel, char *text)
{
   strlwr(kicked);
   if (!strcmp(Nick, kicked))
   {
      Sleep(100);
      IRC.Join(channel);
   }
}

void OnConnect()
{
   StartMessageQueue();
   Connected = TRUE;
   printf("[Connection established.] ");
}

void OnDisconnect()
{
   StopMessageQueue();
   Connected = FALSE;
   printf("[Connection lost.] ");

   if (!ExpectClose && AutoReconnect) TryReconnect = TRUE;
}

void OnError(int err)
{
   printf("[Error: %d  WS: %d] ", err, IRC.GetSocks().LastError());
   if (err == CWE_CLOSE) OnDisconnect();
}

void GetLine(char *line, int len)
{
   int i;

   fgets(line, len, stdin);
   
   i = strlen(line);
   do
   {
      i--;
   }
   while((line[i] == '\n' || line[i] == '\r') && i>=0);
   line[++i] = 0;

   TrimSpace(line);
}

char * StripNick(char *str)
{
   int i,len;
   char *found;

   found = strchr(str, ':');
   if (!found) return NULL;

   len = (int)(found - str);
   for(i=0;i<=len;i++) if ( isspace(str[i])) return NULL;
   for(   ;str[i];i++) if (!isspace(str[i])) break;

   strcpy(str, str+i);

   return str;
}

void TrimSpace(char *str)
{
   int start,end,len;

   len = strlen(str);
   
   for(start=0;start<len;start++) if (!isspace(str[start])) break;
   for(end=len-1;end;end--)       if (!isspace(str[end]))   break;

   if(start) CopyMemory(str, str+start, end-start+1);
   str[end-start+1] = 0;
}

void OnVerbose(char *msg)
{
   PrintMessage(msg);
}

void DoCommaSep(char *str, void (* Funct)(char *))
{
   char *seps  = " ,";
   char *token = strtok(str, seps);

   while(token)
   {
      Funct(token);
      token = strtok(NULL, seps);
   }
}

void AddChan(char *chan)
{
   char *test;
   int i, num = Chans.NumNodes();

   for(i=0;i<num;i++)
   {
      test = Chans.GetDatabyIndex(i);
      if (!_stricmp(chan, test)) return;
   }
   test = new char[strlen(chan)+1];
   strcpy(test, chan);
   Chans.Add(test);
}

void RemChan(char *chan)
{
   char *test;
   int i, num = Chans.NumNodes();

   for(i=0;i<num;i++)
   {
      test = Chans.GetDatabyIndex(i);
      if (!_stricmp(chan, test))
      {
         Chans.DeletebyIndex(i);
         return;
      }
   }
}
