/*  Original Copyright (C) 1998 Jason Hutchens..... heavily modified by Adam Milazzo */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <malloc.h>
#include <string.h>
#include <signal.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include "megahal.h"

#define DWORD unsigned int
// ripped from the windows header files.. if it doesnt compile or link, maybe your headers
// are different from mine
extern "C" __declspec(dllimport) DWORD _stdcall GetTickCount(void);

int ThinkTime;
float SurpriseAvg;
int order=5;
hbool prog=TRUE;
DICTIONARY *ban=NULL;
DICTIONARY *aux=NULL;
DICTIONARY *fin=NULL;
DICTIONARY *grt=NULL;
SWAP *swp=NULL;
FILE *errorfp=NULL;
FILE *statusfp=NULL;
hbool used_key;
char directory[256];
char last[256];

hbool initialize_error(char *filename)
{
	if(errorfp!=NULL) fclose(errorfp);
	if(filename==NULL) return(TRUE);
	errorfp=fopen(filename, "a");
	if(errorfp==NULL) {
		return(FALSE);
	}
	return(TRUE);
}

void error(char *title, char *fmt, ...)
{
	va_list argp;

	if(errorfp!=NULL) {
		fprintf(errorfp, "%s: ", title);
		va_start(argp, fmt);
		vfprintf(errorfp, fmt, argp);
		va_end(argp);
		fprintf(errorfp, ".\n");
		fflush(errorfp);
	} else {
		fprintf(stderr, "%s: ", title);
		va_start(argp, fmt);
		vfprintf(stderr, fmt, argp);
		va_end(argp);
		fprintf(stderr, ".\n");
		fflush(stderr);
	}

	fprintf(stderr, "HalBot died for some reason; check the error log.\n");

	exit(1);
}

hbool warn(char *title, char *fmt, ...)
{
	va_list argp;

	if(errorfp!=NULL) {
		fprintf(errorfp, "%s: ", title);
		va_start(argp, fmt);
		vfprintf(errorfp, fmt, argp);
		va_end(argp);
		fprintf(errorfp, ".\n");
		fflush(errorfp);
	} else {
		fprintf(stderr, "%s: ", title);
		va_start(argp, fmt);
		vfprintf(stderr, fmt, argp);
		va_end(argp);
		fprintf(stderr, ".\n");
		fflush(stderr);
	}

	fprintf(stderr, "HalBot emitted a warning; check the error log.\n");

	return(TRUE);
}

hbool initialize_status(char *filename)
{
	if(statusfp!=NULL) fclose(statusfp);
	if(filename==NULL) return(FALSE);
	statusfp=fopen(filename, "a");
	if(statusfp==NULL) {
		return(FALSE);
	}
	return(TRUE);
}

hbool status(char *fmt, ...)
{
	va_list argp;

	if(statusfp!=NULL) {
		va_start(argp, fmt);
		vfprintf(statusfp, fmt, argp);
		va_end(argp);
		fflush(statusfp);
	} else {
		va_start(argp, fmt);
		vfprintf(stdout, fmt, argp);
		va_end(argp);
		fflush(stdout);
	}

	return(TRUE);
}

void capitalize(char *string)
{
	register int i;
	hbool start=TRUE;

	for(i=0; i<(int)strlen(string); ++i) {
		if(isalpha(string[i])) {
			if(start==TRUE) string[i]=(char)toupper((int)string[i]);
			else string[i]=(char)tolower((int)string[i]);
			start=FALSE;
		}
		if((i>2)&&(strchr("!.?", string[i-1])!=NULL)&&(isspace(string[i])))
			start=TRUE;
	}
}


BYTE2 add_word(DICTIONARY *dictionary, STRING word)
{
	register int i;
	int position;
	hbool found;

	position=search_dictionary(dictionary, word, &found);
	if(found==TRUE) goto succeed;

	dictionary->size+=1;

	if(dictionary->index==NULL) {
		dictionary->index=(BYTE2 *)malloc(sizeof(BYTE2)*
		(dictionary->size));
	} else {
		dictionary->index=(BYTE2 *)realloc((BYTE2 *)
		(dictionary->index),sizeof(BYTE2)*(dictionary->size));
	}
	if(dictionary->index==NULL) {
		error("add_word", "Unable to reallocate the index.");
		goto fail;
	}

	if(dictionary->entry==NULL) {
		dictionary->entry=(STRING *)malloc(sizeof(STRING)*(dictionary->size));
	} else {
		dictionary->entry=(STRING *)realloc((STRING *)(dictionary->entry),
		sizeof(STRING)*(dictionary->size));
	}
	if(dictionary->entry==NULL) {
		error("add_word", "Unable to reallocate the dictionary to %d elements.", dictionary->size);
		goto fail;
	}

	dictionary->entry[dictionary->size-1].length=word.length;
	dictionary->entry[dictionary->size-1].word=(char *)malloc(sizeof(char)*
	(word.length));
	if(dictionary->entry[dictionary->size-1].word==NULL) {
		error("add_word", "Unable to allocate the word.");
		goto fail;
	}
	for(i=0; i<word.length; ++i)
		dictionary->entry[dictionary->size-1].word[i]=word.word[i];

	for(i=(dictionary->size-1); i>position; --i)
		dictionary->index[i]=dictionary->index[i-1];

	dictionary->index[position]=dictionary->size-1;

succeed:
	return(dictionary->index[position]);

fail:
	return(0);
}

int wordcmp(STRING word1, STRING word2)
{
   int ret;

   if (ret = _strnicmp(word1.word, word2.word, MIN(word1.length, word2.length))) return ret;

   if(word1.length<word2.length) return(-1);
   if(word1.length>word2.length) return(1);

   return(0);
}

int search_dictionary(DICTIONARY *dictionary, STRING word, hbool *find)
{
	int position;
	int min;
	int max;
	int middle;
	int compar;

	if(dictionary->size==0) {
		position=0;
		goto notfound;
	}

	min=0;
	max=dictionary->size-1;
	while(TRUE) {
		middle=(min+max)/2;
		compar=wordcmp(word, dictionary->entry[dictionary->index[middle]]);
		if(compar==0) {
			position=middle;
			goto found;
		} else if(compar>0) {
			if(max==middle) {
				position=middle+1;
				goto notfound;
			}
			min=middle+1;
		} else {
			if(min==middle) {
				position=middle;
				goto notfound;
			}
			max=middle-1;
		}
	}

found:
	*find=TRUE;
	return(position);

notfound:
	*find=FALSE;
	return(position);
}

BYTE2 find_word(DICTIONARY *dictionary, STRING word)
{
	int position;
	hbool found;

	position=search_dictionary(dictionary, word, &found);

	if(found==TRUE) return(dictionary->index[position]);
	else return(0);
}

void free_dictionary(DICTIONARY *dictionary)
{
	if(dictionary==NULL) return;
	if(dictionary->entry!=NULL) {
		free(dictionary->entry);
		dictionary->entry=NULL;
	}
	if(dictionary->index!=NULL) {
		free(dictionary->index);
		dictionary->index=NULL;
	}
	dictionary->size=0;
}

void free_model(MODEL *model)
{
	if(model==NULL) return;
	if(model->forward!=NULL) {
		free_tree(model->forward);
	}
	if(model->backward!=NULL) {
		free_tree(model->backward);
	}
	if(model->context!=NULL) {
		free(model->context);
	}
	if(model->dictionary!=NULL) {
		free_dictionary(model->dictionary);
		free(model->dictionary);
	}
	free(model);
}

void free_tree(TREE *tree)
{
	static int level=0;
	register int i;

	if(tree==NULL) return;

	if(tree->tree!=NULL) {
		for(i=0; i<tree->branch; ++i) {
			++level;
			free_tree(tree->tree[i]);
			--level;
		}
		free(tree->tree);
	}
	free(tree);
}

void initialize_dictionary(DICTIONARY *dictionary)
{
	STRING word={ 7, "<ERROR>" };
	STRING end={ 5, "<FIN>" };

	(void)add_word(dictionary, word);
	(void)add_word(dictionary, end);
}

DICTIONARY *new_dictionary(void)
{
	DICTIONARY *dictionary=NULL;

	dictionary=(DICTIONARY *)malloc(sizeof(DICTIONARY));
	if(dictionary==NULL) {
		error("new_dictionary", "Unable to allocate dictionary.");
		return(NULL);
	}

	dictionary->size=0;
	dictionary->index=NULL;
	dictionary->entry=NULL;

	return(dictionary);
}

void save_dictionary(FILE *file, DICTIONARY *dictionary)
{
	register int i;

	fwrite(&(dictionary->size), sizeof(BYTE4), 1, file);
	progress(" Saving dictionary", 0, 1);
	for(i=0; i<dictionary->size; ++i) {
		save_word(file, dictionary->entry[i]);
		progress(NULL, i, dictionary->size);
	}
	progress(NULL, 1, 1);
}

void load_dictionary(FILE *file, DICTIONARY *dictionary)
{
	register int i;
	int size;

	fread(&size, sizeof(BYTE4), 1, file);
	progress(" Loading dictionary", 0, 1);
	for(i=0; i<size; ++i) {
		load_word(file, dictionary);
		progress(NULL, i, size);
	}
	progress(NULL, 1, 1);
}

void save_word(FILE *file, STRING word)
{
	register int i;

	fwrite(&(word.length), sizeof(BYTE1), 1, file);
	for(i=0; i<word.length; ++i)
		fwrite(&(word.word[i]), sizeof(char), 1, file);
}

void load_word(FILE *file, DICTIONARY *dictionary)
{
	register int i;
	STRING word;

	fread(&(word.length), sizeof(BYTE1), 1, file);
	word.word=(char *)malloc(sizeof(char)*word.length);
	if(word.word==NULL) {
		error("load_word", "Unable to allocate word");
		return;
	}
	for(i=0; i<word.length; ++i)
		fread(&(word.word[i]), sizeof(char), 1, file);
	add_word(dictionary, word);
	free(word.word);
}

TREE *new_node(void)
{
	TREE *node=NULL;

	node=(TREE *)malloc(sizeof(TREE));
	if(node==NULL) {
		error("new_node", "Unable to allocate the node.");
		goto fail;
	}

	node->symbol=0;
	node->usage=0;
	node->count=0;
	node->branch=0;
	node->tree=NULL;

	return(node);

fail:
	if(node!=NULL) free(node);
	return(NULL);
}

MODEL *new_model(int order)
{
	MODEL *model=NULL;

	model=(MODEL *)malloc(sizeof(MODEL));
	if(model==NULL) {
		error("new_model", "Unable to allocate model.");
		goto fail;
	}

	model->order=order;
	model->forward=new_node();
	model->backward=new_node();
	model->context=(TREE **)malloc(sizeof(TREE *)*(order+2));
	if(model->context==NULL) {
		error("new_model", "Unable to allocate context array.");
		goto fail;
	}
	initialize_context(model);
	model->dictionary=new_dictionary();
	initialize_dictionary(model->dictionary);

	return(model);

fail:
	return(NULL);
}

void update_model(MODEL *model, int symbol)
{
	register int i;

	for(i=(model->order+1); i>0; --i)
		if(model->context[i-1]!=NULL)
			model->context[i]=add_symbol(model->context[i-1], (BYTE2)symbol);

	return;
}

void update_context(MODEL *model, int symbol)
{
	register int i;

	for(i=(model->order+1); i>0; --i)
		if(model->context[i-1]!=NULL)
			model->context[i]=find_symbol(model->context[i-1], symbol);
}

TREE *add_symbol(TREE *tree, BYTE2 symbol)
{
	TREE *node=NULL;

	node=find_symbol_add(tree, symbol);

	if((node->count<65535)) {
		node->count+=1;
		tree->usage+=1;
	}

	return(node);
}

TREE *find_symbol(TREE *node, int symbol)
{
	register int i;
	TREE *found=NULL;
	hbool found_symbol=FALSE;

	i=search_node(node, symbol, &found_symbol);
	if(found_symbol==TRUE) found=node->tree[i];

	return(found);
}

TREE *find_symbol_add(TREE *node, int symbol)
{
	register int i;
	TREE *found=NULL;
	hbool found_symbol=FALSE;

	i=search_node(node, symbol, &found_symbol);
	if(found_symbol==TRUE) {
		found=node->tree[i];
	} else {
		found=new_node();
		found->symbol=symbol;
		add_node(node, found, i);
	}

	return(found);
}

void add_node(TREE *tree, TREE *node, int position)
{
	register int i;

	if(tree->tree==NULL) {
		tree->tree=(TREE **)malloc(sizeof(TREE *)*(tree->branch+1));
	} else {
		tree->tree=(TREE **)realloc((TREE **)(tree->tree),sizeof(TREE *)*
		(tree->branch+1));
	}
	if(tree->tree==NULL) {
		error("add_node", "Unable to reallocate subtree.");
		return;
	}

	for(i=tree->branch; i>position; --i)
		tree->tree[i]=tree->tree[i-1];

	tree->tree[position]=node;
	tree->branch+=1;
}

int search_node(TREE *node, int symbol, hbool *found_symbol)
{
	register int position;
	int min;
	int max;
	int middle;
	int compar;

	if(node->branch==0) {
		position=0;
		goto notfound;
	}

	min=0;
	max=node->branch-1;
	while(TRUE) {
		middle=(min+max)/2;
		compar=symbol-node->tree[middle]->symbol;
		if(compar==0) {
			position=middle;
			goto found;
		} else if(compar>0) {
			if(max==middle) {
				position=middle+1;
				goto notfound;
			}
			min=middle+1;
		} else {
			if(min==middle) {
				position=middle;
				goto notfound;
			}
			max=middle-1;
		}
	}

found:
	*found_symbol=TRUE;
	return(position);

notfound:
	*found_symbol=FALSE;
	return(position);
}

void initialize_context(MODEL *model)
{
	register int i;

	for(i=0; i<=model->order; ++i) model->context[i]=NULL;
}

void learn(MODEL *model, DICTIONARY *words)
{
	register int i;
	BYTE2 symbol;

	if(words->size<=(model->order)) return;

	initialize_context(model);
	model->context[0]=model->forward;
	for(i=0; i<words->size; ++i) {
		symbol=add_word(model->dictionary, words->entry[i]);
		update_model(model, symbol);
	}
	update_model(model, 1);

	initialize_context(model);
	model->context[0]=model->backward;
	for(i=words->size-1; i>=0; --i) {
		symbol=find_word(model->dictionary, words->entry[i]);
		update_model(model, symbol);
	}
	update_model(model, 1);

	return;
}

int train(MODEL *model, char *filename)
{
	FILE *file;
	char buffer[1024];
	DICTIONARY *words=NULL;
	int length;

	if(filename==NULL) return 1;

	file=fopen(filename, "r");
	if(file==NULL) {
		printf(" Unable to open \"%s\".\n", filename);
		return 1;
	}

	fseek(file, 0, 2);
   length=ftell(file);
   rewind(file);

	words=new_dictionary();

	progress(" Training from file", 0, 1);
	while(!feof(file)) {

		if(fgets(buffer, 1024, file)==NULL) break;
		if(buffer[0]=='#') continue;

		buffer[strlen(buffer)-1]='\0';

		strupr(buffer);
		make_words(buffer, words);
		learn(model, words);

		progress(NULL, ftell(file), length);

	}
	progress(NULL, 1, 1);

	free_dictionary(words);
	fclose(file);

   return 0;
}

void save_model(char *modelname, MODEL *model)
{
	FILE *file;
	static char *filename=NULL;
	
	if(filename==NULL) filename=(char *)malloc(sizeof(char)*1);

	filename=(char *)realloc(filename,
		sizeof(char)*(strlen(directory)+strlen(SEP)+12));
	if(filename==NULL)
   {
      error("save_model","Unable to allocate filename");
      return;
   }

	sprintf(filename, "%s%s%s", directory, SEP, modelname);
	file=fopen(filename, "wb");
	if(file==NULL) {
		warn("save_model", "Unable to open file `%s'", filename);
		return;
	}

	fwrite(COOKIE, sizeof(char), strlen(COOKIE), file);
	fwrite(&(model->order), sizeof(BYTE1), 1, file);
	save_tree(file, model->forward);
	save_tree(file, model->backward);
	save_dictionary(file, model->dictionary);

	fclose(file);
}

void save_tree(FILE *file, TREE *node)
{
	static int level=0;
	register int i;

	fwrite(&(node->symbol), sizeof(BYTE2), 1, file);
	fwrite(&(node->usage), sizeof(BYTE4), 1, file);
	fwrite(&(node->count), sizeof(BYTE2), 1, file);
	fwrite(&(node->branch), sizeof(BYTE2), 1, file);

	if(level==0) progress(" Saving tree", 0, 1);
	for(i=0; i<node->branch; ++i) {
		++level;
		save_tree(file, node->tree[i]);
		--level;
		if(level==0) progress(NULL, i, node->branch);
	}
	if(level==0) progress(NULL, 1, 1);
}

void load_tree(FILE *file, TREE *node)
{
	static int level=0;
	register int i;

	fread(&(node->symbol), sizeof(BYTE2), 1, file);
	fread(&(node->usage), sizeof(BYTE4), 1, file);
	fread(&(node->count), sizeof(BYTE2), 1, file);
	fread(&(node->branch), sizeof(BYTE2), 1, file);

	if(node->branch==0) return;

	node->tree=(TREE **)malloc(sizeof(TREE *)*(node->branch));
	if(node->tree==NULL) {
		error("load_tree", "Unable to allocate subtree");
		return;
	}

	if(level==0) progress(" Loading tree", 0, 1);
	for(i=0; i<node->branch; ++i) {
		node->tree[i]=new_node();
		++level;
		load_tree(file, node->tree[i]);
		--level;
		if(level==0) progress(NULL, i, node->branch);
	}
	if(level==0) progress(NULL, 1, 1);
}

hbool load_model(char *fn, MODEL **model, DICTIONARY **b, DICTIONARY **a,
                 DICTIONARY **g, SWAP **s, FILE **trn)
{
	FILE *file;
	char cookie[16];
   char filename[256];

   *model=new_model(order);

	if(fn==NULL)
   {
	   *b=initialize_list(NULL);
	   *a=initialize_list(NULL);
	   *g=initialize_list(NULL);
	   *s=initialize_swap(NULL);
      return(TRUE);
   }

	sprintf(filename, "%s%s%s.ban", directory, SEP, fn);
	*b=initialize_list(filename);
	sprintf(filename, "%s%s%s.aux", directory, SEP, fn);
	*a=initialize_list(filename);
	sprintf(filename, "%s%s%s.grt", directory, SEP, fn);
	*g=initialize_list(filename);
	sprintf(filename, "%s%s%s.swp", directory, SEP, fn);
	*s=initialize_swap(filename);
   sprintf(filename, "%s%s%s.trn", directory, SEP, fn);
   *trn = fopen(filename, "a+");

	if(strcmp(directory, DEFAULT)!=0)
   {
	   sprintf(filename, "%s%s%s.brn", directory, SEP, fn);
	   file=fopen(filename, "r");
	   if(file==NULL)
      {
		   sprintf(filename, "%s%s%s.brn", directory, SEP, fn);
		   file=fopen(filename, "r");
		   if(file==NULL)
         {
			   fprintf(stdout, "Unable to change HalBot personality to \"%s\".\n"
				   "Reverting to HalBot personality \"%s\".\n", directory, last);
			   strcpy(directory,strdup(last));
			   return FALSE;
		   }
	   }

	fclose(file);
	fprintf(stdout, "Changing to HalBot personality \"%s\".\n", directory);
	}
   else sprintf(filename, "%s.brn", fn);

   file=fopen(filename, "rb");
	if(file==NULL) {
		warn("load_model", "Unable to open file `%s'", filename);
		return(FALSE);
	}

	fread(cookie, sizeof(char), strlen(COOKIE), file);
	if(strncmp(cookie, COOKIE, strlen(COOKIE))!=0) {
		warn("load_model", "File `%s' is not a HalBot brain", filename);
		goto fail;
	}

	fread(&(*model)->order, sizeof(BYTE1), 1, file);
	load_tree(file, (*model)->forward);
	load_tree(file, (*model)->backward);
	load_dictionary(file, (*model)->dictionary);


	return(TRUE);
fail:
	fclose(file);

	return(FALSE);
}

void make_words(char *input, DICTIONARY *words)
{
	int offset=0;

	free_dictionary(words);

	if(strlen(input)==0) return;

	while(1) {

		if(boundary(input, offset)) {
			if(words->entry==NULL)
				words->entry=(STRING *)malloc((words->size+1)*sizeof(STRING));
			else
				words->entry=(STRING *)realloc(words->entry, (words->size+1)*sizeof(STRING));

			if(words->entry==NULL) {
				error("make_words", "Unable to reallocate dictionary");
				return;
			}

			words->entry[words->size].length=offset;
			words->entry[words->size].word=input;
			words->size+=1;

			if(offset==(int)strlen(input)) break;
			input+=offset;
			offset=0;
		} else {
			++offset;
		}
	}

	if(isalnum(words->entry[words->size-1].word[0])) {
		if(words->entry==NULL)
			words->entry=(STRING *)malloc((words->size+1)*sizeof(STRING));
		else
			words->entry=(STRING *)realloc(words->entry, (words->size+1)*sizeof(STRING));
		if(words->entry==NULL) {
			error("make_words", "Unable to reallocate dictionary");
			return;
		}

		words->entry[words->size].length=1;
		words->entry[words->size].word=".";
		++words->size;
	}
	else if(strchr("!.?", words->entry[words->size-1].word[words->entry[words->size-1].length-1])==NULL) {
		words->entry[words->size-1].length=1;
		words->entry[words->size-1].word=".";
	}

   return;
}
 
hbool boundary(char *string, int position)
{
	if(position==0)
		return(FALSE);

	if(position==(int)strlen(string))
		return(TRUE);

	if(
		(string[position]=='\'')&&
		(isalpha(string[position-1])!=0)&&
		(isalpha(string[position+1])!=0)
	)
		return(FALSE);

	if(
		(position>1)&&
		(string[position-1]=='\'')&&
		(isalpha(string[position-2])!=0)&&
		(isalpha(string[position])!=0)
	)
		return(FALSE);

	if(
		(isalpha(string[position])!=0)&&
		(isalpha(string[position-1])==0)
	)
		return(TRUE);
	
	if(
		(isalpha(string[position])==0)&&
		(isalpha(string[position-1])!=0)
	)
		return(TRUE);
	
	if(isdigit(string[position])!=isdigit(string[position-1]))
		return(TRUE);

	return(FALSE);
}
 
char *generate_reply(MODEL *model, DICTIONARY *words)
{
	static DICTIONARY *dummy=NULL;
	DICTIONARY *replywords;
	DICTIONARY *keywords;
	float diff, min_diff;
   float strive = SurpriseAvg;
	char *output;
	static char *output_none=NULL;
	DWORD basetime, thinktime = ThinkTime;

	keywords=make_keywords(model, words);

	output=output_none;
	if(dummy==NULL) dummy=new_dictionary();
	replywords=reply(model, dummy);
	if(dissimilar(words, replywords)==TRUE) output=make_output(replywords);

	min_diff = 9999999.0;
	basetime=GetTickCount();

	do {
		replywords=reply(model, keywords);
      diff = fabs(strive - evaluate_reply(model, keywords, replywords));
      if((diff<min_diff)&&(dissimilar(words, replywords)==TRUE)) {
			min_diff=diff;
			output=make_output(replywords);
		}
	} while((GetTickCount()-basetime)<thinktime);
	return(output);
}

hbool dissimilar(DICTIONARY *words1, DICTIONARY *words2)
{
	register int i;

	if(words1->size!=words2->size) return(TRUE);
	for(i=0; i<words1->size; ++i)
		if(wordcmp(words1->entry[i], words2->entry[i])!=0) return(TRUE);
	return(FALSE);
}

DICTIONARY *make_keywords(MODEL *model, DICTIONARY *words)
{
	static DICTIONARY *keys=NULL;
	register int i;
	register int j;
	int c;

	if(keys==NULL) keys=new_dictionary();
	for(i=0; i<keys->size; ++i) free(keys->entry[i].word);
	free_dictionary(keys);

	for(i=0; i<words->size; ++i) {
		c=0;
		for(j=0; j<swp->size; ++j)
			if(wordcmp(swp->from[j], words->entry[i])==0) {
				add_key(model, keys, swp->to[j]);
				++c;
			}
		if(c==0) add_key(model, keys, words->entry[i]);
	}

	if(keys->size>0) for(i=0; i<words->size; ++i) {

		c=0;
		for(j=0; j<swp->size; ++j)
			if(wordcmp(swp->from[j], words->entry[i])==0) {
				add_aux(model, keys, swp->to[j]);
				++c;
			}
		if(c==0) add_aux(model, keys, words->entry[i]);
	}

	return(keys);
}

void add_key(MODEL *model, DICTIONARY *keys, STRING word)
{
	int symbol;

	symbol=find_word(model->dictionary, word);
	if(symbol==0) return;
	if(isalnum(word.word[0])==0) return;
	symbol=find_word(ban, word);
	if(symbol!=0) return;
	symbol=find_word(aux, word);
	if(symbol!=0) return;

	add_word(keys, word);
}

void add_aux(MODEL *model, DICTIONARY *keys, STRING word)
{
	int symbol;

	symbol=find_word(model->dictionary, word);
	if(symbol==0) return;
	if(isalnum(word.word[0])==0) return;
	symbol=find_word(aux, word);
	if(symbol==0) return;

	add_word(keys, word);
}

DICTIONARY *reply(MODEL *model, DICTIONARY *keys)
{
	static DICTIONARY *replies=NULL;
	register int i;
	int symbol;
	hbool start=TRUE;

	if(replies==NULL) replies=new_dictionary();
	free_dictionary(replies);

	initialize_context(model);
	model->context[0]=model->forward;
	used_key=FALSE;

	while(TRUE) {
		if(start==TRUE) symbol=seed(model, keys);
		else symbol=babble(model, keys, replies);
		if((symbol==0)||(symbol==1)) break;
		start=FALSE;

		if(replies->entry==NULL)
			replies->entry=(STRING *)malloc((replies->size+1)*sizeof(STRING));
		else
			replies->entry=(STRING *)realloc(replies->entry, (replies->size+1)*sizeof(STRING));
		if(replies->entry==NULL) {
			error("reply", "Unable to reallocate dictionary");
			return(NULL);
		}

		replies->entry[replies->size].length=
			model->dictionary->entry[symbol].length;
		replies->entry[replies->size].word=
			model->dictionary->entry[symbol].word;
		replies->size+=1;

		update_context(model, symbol);
	}

	initialize_context(model);
	model->context[0]=model->backward;

	if(replies->size>0) for(i=MIN(replies->size-1, model->order); i>=0; --i) {
		symbol=find_word(model->dictionary, replies->entry[i]);
		update_context(model, symbol);
	}

	while(TRUE) {
		symbol=babble(model, keys, replies);
		if((symbol==0)||(symbol==1)) break;

		if(replies->entry==NULL)
			replies->entry=(STRING *)malloc((replies->size+1)*sizeof(STRING));
		else
			replies->entry=(STRING *)realloc(replies->entry, (replies->size+1)*sizeof(STRING));
		if(replies->entry==NULL) {
			error("reply", "Unable to reallocate dictionary");
			return(NULL);
		}

		for(i=replies->size; i>0; --i) {
			replies->entry[i].length=replies->entry[i-1].length;
			replies->entry[i].word=replies->entry[i-1].word;
		}

		replies->entry[0].length=model->dictionary->entry[symbol].length;
		replies->entry[0].word=model->dictionary->entry[symbol].word;
		replies->size+=1;

		update_context(model, symbol);
	}

	return(replies);
}

float evaluate_reply(MODEL *model, DICTIONARY *keys, DICTIONARY *words)
{
	register int i;
	register int j;
	int symbol;
	float probability;
	int count;
	float entropy=(float)0.0;
	TREE *node;
	int num=0;

	if(words->size<=0) return((float)0.0);
	initialize_context(model);
	model->context[0]=model->forward;
	for(i=0; i<words->size; ++i) {
		symbol=find_word(model->dictionary, words->entry[i]);

		if(find_word(keys, words->entry[i])!=0) {
			probability=(float)0.0;
			count=0;
			++num;
			for(j=0; j<model->order; ++j) if(model->context[j]!=NULL) {
	
				node=find_symbol(model->context[j], symbol);
				probability+=(float)(node->count)/
					(float)(model->context[j]->usage);
				++count;
	
			}

			if(count>0) entropy-=(float)log(probability/(float)count);
		}

		update_context(model, symbol);
	}

	initialize_context(model);
	model->context[0]=model->backward;
	for(i=words->size-1; i>=0; --i) {
		symbol=find_word(model->dictionary, words->entry[i]);

		if(find_word(keys, words->entry[i])!=0) {
			probability=(float)0.0;
			count=0;
			++num;
			for(j=0; j<model->order; ++j) if(model->context[j]!=NULL) {
	
				node=find_symbol(model->context[j], symbol);
				probability+=(float)(node->count)/
					(float)(model->context[j]->usage);
				++count;
	
			}

			if(count>0) entropy-=(float)log(probability/(float)count);
		}

		update_context(model, symbol);
	}

	if(num>=8) entropy/=(float)sqrt(num-1);
	if(num>=16) entropy/=(float)num;

	return(entropy);
}

char *make_output(DICTIONARY *words)
{
	static char *output=NULL;
	register int i;
	register int j;
	int length;
	static char *output_none=NULL;
	
	if(output_none==NULL) output_none=(char *)malloc(40);

	if(output==NULL) {
		output=(char *)malloc(sizeof(char));
		if(output==NULL) {
			error("make_output", "Unable to allocate output");
			return(output_none);
		}
	}

	if(words->size==0) {
		return(NULL);
	}

	length=1;
	for(i=0; i<words->size; ++i) length+=words->entry[i].length;

	output=(char *)realloc(output, sizeof(char)*length);
	if(output==NULL) {
		error("make_output", "Unable to reallocate output.");
		return(NULL);
	}

	length=0;
	for(i=0; i<words->size; ++i)
		for(j=0; j<words->entry[i].length; ++j)
			output[length++]=words->entry[i].word[j];
			
	output[length]='\0';

	return(output);
}

int babble(MODEL *model, DICTIONARY *keys, DICTIONARY *words)
{
	TREE *node;
	register int i;
	int count;
	int symbol;

	for(i=0; i<=model->order; ++i)
		if(model->context[i]!=NULL)
			node=model->context[i];

	if(node->branch==0) return(0);

	i=rnd(node->branch);
	count=rnd(node->usage);
	while(count>=0) {
		symbol=node->tree[i]->symbol;

		if(
			(find_word(keys, model->dictionary->entry[symbol])!=0)&&
			((used_key==TRUE)||
			(find_word(aux, model->dictionary->entry[symbol])==0))&&
			(word_exists(words, model->dictionary->entry[symbol])==FALSE)
		) {
			used_key=TRUE;
			break;
		}
		count-=node->tree[i]->count;
		i=(i>=(node->branch-1))?0:i+1;
	}

	return(symbol);
}

hbool word_exists(DICTIONARY *dictionary, STRING word)
{
	register int i;

	for(i=0; i<dictionary->size; ++i)
		if(wordcmp(dictionary->entry[i], word)==0)
			return(TRUE);
	return(FALSE);
}

int seed(MODEL *model, DICTIONARY *keys)
{
	register int i;
	int symbol;
	int stop;

	if(model->context[0]->branch==0) symbol=0;
	else symbol=model->context[0]->tree[rnd(model->context[0]->branch)]->symbol;

	if(keys->size>0) {
		i=rnd(keys->size);
		stop=i;
		while(TRUE) {
			if(
				(find_word(model->dictionary, keys->entry[i])!=0)&&
				(find_word(aux, keys->entry[i])==0)
			) {
				symbol=find_word(model->dictionary, keys->entry[i]);
				return(symbol);
			}
			++i;
			if(i==keys->size) i=0;
			if(i==stop) return(symbol);
		}
	}

	return(symbol);
}

SWAP *new_swap(void)
{
	SWAP *list;

	list=(SWAP *)malloc(sizeof(SWAP));
	if(list==NULL) {
		error("new_swap", "Unable to allocate swap");
		return(NULL);
	}
	list->size=0;
	list->from=NULL;
	list->to=NULL;

	return(list);
}

void add_swap(SWAP *list, char *s, char *d)
{
	list->size+=1;

	if(list->from==NULL) {
		list->from=(STRING *)malloc(sizeof(STRING));
		if(list->from==NULL) {
			error("add_swap", "Unable to allocate list->from");
			return;
		}
	}

	if(list->to==NULL) {
		list->to=(STRING *)malloc(sizeof(STRING));
		if(list->to==NULL) {
			error("add_swap", "Unable to allocate list->to");
			return;
		}
	}

	list->from=(STRING *)realloc(list->from, sizeof(STRING)*(list->size));
	if(list->from==NULL) {
		error("add_swap", "Unable to reallocate from");
		return;
	}

	list->to=(STRING *)realloc(list->to, sizeof(STRING)*(list->size));
	if(list->to==NULL) {
		error("add_swap", "Unable to reallocate to");
		return;
	}

	list->from[list->size-1].length=strlen(s);
	list->from[list->size-1].word=strdup(s);
	list->to[list->size-1].length=strlen(d);
	list->to[list->size-1].word=strdup(d);
}

SWAP *initialize_swap(char *filename)
{
	SWAP *list;
	FILE *file=NULL;
	char buffer[1024];
	char *from;
	char *to;

	list=new_swap();

	if(filename==NULL) return(list);

	file=fopen(filename, "r");
	if(file==NULL) return(list);

	while(!feof(file)) {

		if(fgets(buffer, 1024, file)==NULL) break;
		if(buffer[0]=='#') continue;
		from=strtok(buffer, "\t ");
		to=strtok(NULL, "\t \n#");

		if (from && to) add_swap(list, from, to);
	}

	fclose(file);
	return(list);
}

void free_swap(SWAP *swap)
{
	register int i;

	if(swap==NULL) return;

	for(i=0; i<swap->size; ++i) {
		free_word(swap->from[i]);
		free_word(swap->to[i]);
	}
	free(swap->from);
	free(swap->to);
	free(swap);
}

DICTIONARY *initialize_list(char *filename)
{
	DICTIONARY *list;
	FILE *file=NULL;
	STRING word;
	char *string;
	char buffer[1024];

	list=new_dictionary();

	if(filename==NULL) return(list);

	file=fopen(filename, "r");
	if(file==NULL) return(list);

	while(!feof(file)) {

		if(fgets(buffer, 1024, file)==NULL) break;
		if(buffer[0]=='#') continue;
		string=strtok(buffer, "\t \n#");

		if((string!=NULL)&&(strlen(string)>0)) {
			word.length=strlen(string);
			word.word=strdup(buffer);
			add_word(list, word);
		}
	}

	fclose(file);
	return(list);
}

int rnd(int range)
{
	static hbool flag=FALSE;

	if(flag==FALSE) {
		srand(time(NULL));
	}
	flag=TRUE;
	return(rand()%range);
}

hbool progress(char *message, int done, int total)
{
	static int last=0;
	static hbool first=FALSE;

	if(prog==FALSE) return(TRUE);
 
	if((done*100/total==100)&&(first==FALSE)) return(TRUE);

	if(done*100/total==last) {
		if((done==0)&&(first==FALSE)) {
			fprintf(stderr, "%s: %3d%%", message, done*100/total);
			first=TRUE;
		}
		return(TRUE);
	}

	last=done*100/total;

	if(done>0) fprintf(stderr, "%c%c%c%c", 8, 8, 8, 8);
	fprintf(stderr, "%3d%%", done*100/total);

	if(last==100) {
		first=FALSE;
		last=0;
		fprintf(stderr, "\n");
	}

	return(TRUE);
}

void load_personality(DICTIONARY *b, DICTIONARY *a, DICTIONARY *g, SWAP *s)
{
   ban = b;
   aux = a;
   grt = g;
   swp = s;

}

void change_personality(DICTIONARY *command, int position)
{
	last[0] = 0;
	
   if(directory[0]) strcpy(last,directory);
	if((command==NULL)||((position+2)>=command->size)) {
		strcpy(directory, DEFAULT);
		if(!last[0]) strcpy(last,directory);
	} else {
		strncpy(directory, command->entry[position+2].word,
			command->entry[position+2].length);
		directory[command->entry[position+2].length]='\0';
	}
}

void free_words(DICTIONARY *words)
{
	register int i;

	if(words==NULL) return;

	if(words->entry!=NULL)
		for(i=0; i<words->size; ++i) free_word(words->entry[i]);
}

void free_word(STRING word)
{
	free(word.word);
}

