/*  Original Copyright (C) 1998 Jason Hutchens..... heavily modified by Adam Milazzo */

#ifndef _MEGAHAL_H
#define _MEGAHAL_H

#include <stdio.h>

#define MIN(a,b) ((a)<(b))?(a):(b)

#define COOKIE "MegaHALv8"

#define DEFAULT "."

#define BYTE1 unsigned char
#define BYTE2 unsigned short
#define BYTE4 unsigned long

#define SEP "\\"

#undef FALSE
#undef TRUE
typedef enum { FALSE, TRUE } hbool;

typedef struct {
	BYTE1 length;
	char *word;
} STRING;

typedef struct {
	BYTE4 size;
	STRING *entry;
	BYTE2 *index;
} DICTIONARY;

typedef struct {
	BYTE2 size;
	STRING *from;
	STRING *to;
} SWAP;

typedef struct NODE {
	BYTE2 symbol;
	BYTE4 usage;
	BYTE2 count;
	BYTE2 branch;
	struct NODE **tree;
} TREE;

typedef struct {
	BYTE1 order;
	TREE *forward;
	TREE *backward;
	TREE **context;
	DICTIONARY *dictionary;
} MODEL;

void add_aux(MODEL *, DICTIONARY *, STRING);
void add_key(MODEL *, DICTIONARY *, STRING);
void add_node(TREE *, TREE *, int);
void add_swap(SWAP *, char *, char *);
TREE *add_symbol(TREE *, BYTE2);
BYTE2 add_word(DICTIONARY *, STRING);
int babble(MODEL *, DICTIONARY *, DICTIONARY *);
hbool boundary(char *, int);
void capitalize(char *);
void change_personality(DICTIONARY *command, int position);
void change_think(DICTIONARY *, int);
hbool dissimilar(DICTIONARY *, DICTIONARY *);
void error(char *, char *, ...);
float evaluate_reply(MODEL *, DICTIONARY *, DICTIONARY *);
TREE *find_symbol(TREE *, int);
TREE *find_symbol_add(TREE *, int);
BYTE2 find_word(DICTIONARY *, STRING);
void free_dictionary(DICTIONARY *);
void free_model(MODEL *);
void free_tree(TREE *);
void free_word(STRING);
void free_words(DICTIONARY *);
char *generate_reply(MODEL *, DICTIONARY *);
void initialize_context(MODEL *);
void initialize_dictionary(DICTIONARY *);
hbool initialize_error(char *);
DICTIONARY *initialize_list(char *);
hbool initialize_status(char *);
SWAP *initialize_swap(char *);
void learn(MODEL *, DICTIONARY *);
void load_dictionary(FILE *, DICTIONARY *);
hbool load_model(char *fn, MODEL **model, DICTIONARY **b, DICTIONARY **a,
                 DICTIONARY **g, SWAP **s, FILE **trn);
void load_personality(DICTIONARY *b, DICTIONARY *a, DICTIONARY *g, SWAP *s);
void load_tree(FILE *, TREE *);
void load_word(FILE *, DICTIONARY *);
DICTIONARY *make_keywords(MODEL *, DICTIONARY *);
char *make_output(DICTIONARY *);
void make_words(char *, DICTIONARY *);
DICTIONARY *new_dictionary(void);
MODEL *new_model(int);
TREE *new_node(void);
SWAP *new_swap(void);
hbool progress(char *, int, int);
DICTIONARY *reply(MODEL *, DICTIONARY *);
void save_dictionary(FILE *, DICTIONARY *);
void save_model(char *, MODEL *);
void save_tree(FILE *, TREE *);
void save_word(FILE *, STRING);
int search_dictionary(DICTIONARY *, STRING, hbool *);
int search_node(TREE *, int, hbool *);
int seed(MODEL *, DICTIONARY *);
hbool status(char *, ...);
int train(MODEL *, char *);
void update_context(MODEL *, int);
void update_model(MODEL *, int);
hbool warn(char *, char *, ...);
int wordcmp(STRING word1, STRING word2);
hbool word_exists(DICTIONARY *, STRING);
void write_input(char *);
void write_output(char *);
int rnd(int);

#endif
