#include "CIRC.h"
#include <windows.h>
#include <stdio.h>
#include "MsgQueue.h"
#include "LinkList.h"

struct   Message
{
   DWORD Time;
   int   MsgLen;
   int   DestLen;
   int   ID;
   char *Msg;
   char *Dest;
};

struct   Print
{
   char *Msg;
   char *Dest;
   int   MsgLen;
   int   DestLen;
   int   ID;
};

static   HANDLE            hQueueThread;
static   CRITICAL_SECTION  MQ_CS;
static   Message          *Msgs            = NULL;
static   Print            *Prints          = NULL;
static   volatile int      MaxMessages     = 0;
static   volatile int      MaxPrints       = 0;
static   volatile int      MessagesWaiting = 0;
static   volatile int      PrintsWaiting   = 0;
static   volatile int      Serial          = 1;
static   volatile int      nPID            = 1;    // next ID to be assigned
static   volatile int      fPID            = 1;    // next ID to be printed
static   volatile bool     KillThread      = FALSE;
static   volatile bool     QueueRunning    = FALSE;

extern   CIRC              IRC;
extern   volatile bool     HandlingInput;

static   DWORD WINAPI      QueueFunc(LPVOID ignored);


int InitMessageQueue()
{
   SECURITY_ATTRIBUTES sa;
   DWORD id;

   if (hQueueThread == NULL)
   {   
      sa.nLength = sizeof(SECURITY_ATTRIBUTES);
      sa.lpSecurityDescriptor = NULL;
      sa.bInheritHandle = FALSE;

      hQueueThread = CreateThread(&sa, 0, QueueFunc, NULL, 0, &id);
      if (hQueueThread == NULL) return 1;
   }

   InitializeCriticalSection(&MQ_CS);
   return 0;
}

void DeinitMessageQueue()
{
   StopMessageQueue();
   ClearMessageQueue();

   KillThread = TRUE;
   WaitForSingleObject(hQueueThread, INFINITE);
   CloseHandle(hQueueThread);
   hQueueThread = NULL;
}

void StartMessageQueue()
{
   EnterCriticalSection(&MQ_CS);
   QueueRunning = TRUE;
   LeaveCriticalSection(&MQ_CS);
}

void StopMessageQueue()
{
   EnterCriticalSection(&MQ_CS);
   QueueRunning = FALSE;
   LeaveCriticalSection(&MQ_CS);
}

void ClearMessageQueue()
{
   int   i;

   EnterCriticalSection(&MQ_CS);
   for(i=0;i<MaxMessages;i++) Msgs[i].ID = 0;
   LeaveCriticalSection(&MQ_CS);
}

void FlushMessageQueue()
{
   int   i, j;

   EnterCriticalSection(&MQ_CS);
   
   for(i=0;i<MaxMessages;i++)
   {
      if (!Msgs[i].ID) continue;
      for(j=0;j<MaxMessages;j++)
      {
         if (!Msgs[j].ID || i == j) continue;
         if (!strcmp(Msgs[i].Dest,Msgs[j].Dest) && Msgs[i].ID > Msgs[j].ID) goto Skip;
      }
      IRC.SendMessage(Msgs[i].Dest, Msgs[i].Msg);
      Msgs[i].ID = 0;
      Skip:
      ; // empty statement to avoid an error :)
   }
   MessagesWaiting = 0;

   LeaveCriticalSection(&MQ_CS);
}
   
void StartMessage(char *dest, char *str, int delay)
{
   int i, j, len;
   DWORD curtime = GetTickCount();
   static char *MemErr = "Error allocating memory for StartMessage.";

   EnterCriticalSection(&MQ_CS);

   if(!Msgs) // start a new list
   {
      Msgs = (Message *)calloc(sizeof(Message), MaxMessages = 50);
      if (!Msgs)
      {
         puts(MemErr);
         goto End;
      }
      i = 0;
   }
   else     // find a free message
   {
      for(i=0;i<MaxMessages && Msgs[i].ID;i++);
      if (i == MaxMessages) // no free messages.. allocate space for some more
      {
         Msgs = (Message *)realloc(Msgs, (MaxMessages + 25) * sizeof(Message));
         if (!Msgs)
         {
            puts(MemErr);
            goto End;
         }
         ZeroMemory(Msgs + MaxMessages, 25*sizeof(Message));
         MaxMessages += 25;
      }
   }
   // i points to where we will store our message

   len = strlen(str);
   if (Msgs[i].MsgLen <= len) // not enough space for our message
   {
      Msgs[i].Msg = (char *)realloc(Msgs[i].Msg, len+64);
      if (!Msgs[i].Msg)
      {
         puts(MemErr);
         goto End;
      }
      Msgs[i].MsgLen = len+64;
   }
   CopyMemory(Msgs[i].Msg, str, len+1);

   len = strlen(dest);
   if (Msgs[i].DestLen <= len) // not enough space for our dest
   {
      Msgs[i].Dest = (char *)realloc(Msgs[i].Dest, len+32);
      if (!Msgs[i].Dest)
      {
         puts(MemErr);
         goto End;
      }
      Msgs[i].DestLen = len+32;
   }
   CopyMemory(Msgs[i].Dest, dest, len+1);
   
   for(j=0;j<MaxMessages;j++)    // add to the delay if 
   {
      if (!Msgs[j].ID) continue;
      if (!strcmp(Msgs[i].Dest, Msgs[j].Dest))
      {
         if (Msgs[j].Time > curtime) curtime = Msgs[j].Time;
      }
   }

   Msgs[i].ID   = Serial++;
   if (delay < 0) delay = 0;
   Msgs[i].Time = curtime + delay;
   MessagesWaiting++;

   End:
   LeaveCriticalSection(&MQ_CS);
}


void  PrintMessage(char *msg, char *dest)
{
   int i, len;
   static char *MemErr = "Error allocating memory for PrintMessage.";
   
   EnterCriticalSection(&MQ_CS);

   if(!Prints) // start a new list
   {
      Prints = (Print *)calloc(sizeof(Print), MaxPrints = 20);
      if (!Prints)
      {
         puts(MemErr);
         goto End;
      }
      i = 0;
   }
   else     // find a free print
   {
      for(i=0;i<MaxPrints && Prints[i].ID;i++);
      if (i == MaxPrints) // no free prints.. allocate space for some more
      {
         Prints = (Print *)realloc(Prints, (MaxPrints + 15) * sizeof(Print));
         if (!Prints)
         {
            puts(MemErr);
            goto End;
         }
         ZeroMemory(Prints + MaxPrints, 15*sizeof(Print));
         MaxPrints += 15;
      }
   }
   // i points to where we will store our print

   len = strlen(msg);
   if (Prints[i].MsgLen <= len) // not enough space for our message
   {
      Prints[i].Msg = (char *)realloc(Prints[i].Msg, len+96);
      if (!Prints[i].Msg)
      {
         puts(MemErr);
         goto End;
      }
      Prints[i].MsgLen = len+96;
   }
   CopyMemory(Prints[i].Msg, msg, len+1);
   if(dest)
   {
      len = strlen(dest);
      if(Prints[i].DestLen <= len)
      {
         Prints[i].Dest = (char *)realloc(Prints[i].Dest, len+32);
         if(!Prints[i].Msg)
         {
            puts(MemErr);
            goto End;
         }
         Prints[i].MsgLen = len+32;
      }
      CopyMemory(Prints[i].Dest, dest, len+1);
   }
   else
   {
      if(Prints[i].DestLen) Prints[i].Dest[0]=0;
      else Prints[i].Dest = (char *)calloc(1, 32), Prints[i].DestLen = 32;
   }

   Prints[i].ID   = nPID++;
   PrintsWaiting++;

   End:
   LeaveCriticalSection(&MQ_CS);
}
   

void PrintFMessage(char *ctrl, ...)
{
   va_list list;

   va_start(list, ctrl);
   VPrintMessage(ctrl, list);
   va_end(list);
}


void VPrintMessage(char *ctrl, va_list list)
{
   char buf[2048];

   vsprintf(buf, ctrl, list);
   PrintMessage(buf);
}


DWORD WINAPI QueueFunc(LPVOID ignored)
{
   int i, j;
   DWORD curtime;

   while(!KillThread)
   {
      if (PrintsWaiting && !HandlingInput)
      {
         EnterCriticalSection(&MQ_CS);
         while(PrintsWaiting)
         {
            for(i=0;i<MaxPrints;i++)
            {
               if (Prints[i].ID != fPID) continue;
               fPID++;
               if(Prints[i].Dest[0]) IRC.SendMessage(Prints[i].Dest, Prints[i].Msg);
               else puts(Prints[i].Msg);
               Prints[i].ID = 0;
               if (! --PrintsWaiting) break;
            }
         }
         LeaveCriticalSection(&MQ_CS);
      }
            
      if (MessagesWaiting && QueueRunning)
      {
         EnterCriticalSection(&MQ_CS);
         curtime = GetTickCount();
         for(i=0;i<MaxMessages;i++)
         {
            if (!Msgs[i].ID) continue;
            if (curtime > Msgs[i].Time)
            {
               for(j=0;j<MaxMessages;j++)
               {
                  if (!Msgs[j].ID || i == j) continue;
                  if (!strcmp(Msgs[i].Dest,Msgs[j].Dest) && Msgs[i].ID > Msgs[j].ID) goto Skip;
               }
               IRC.SendMessage(Msgs[i].Dest, Msgs[i].Msg);
               Msgs[i].ID = 0;
               if (--MessagesWaiting == 0) break;
               Skip:
               ; // empty statement to avoid an error :)
            }
         }
         LeaveCriticalSection(&MQ_CS);
      }
      Sleep(50);
   }

   ExitThread(0);
   return 0;
}
